"""Algorithm for replacing form arguments with 'stripped' versions.

In the stripped version, any data-carrying objects have been extracted to a mapping.
"""

from ufl.algorithms.replace import replace
from ufl.classes import (
    Argument,
    Coefficient,
    Constant,
    Form,
    FunctionSpace,
    Integral,
    Mesh,
    MeshView,
    MixedFunctionSpace,
    TensorProductFunctionSpace,
)
from ufl.corealg.map_dag import map_expr_dag
from ufl.corealg.multifunction import MultiFunction


class TerminalStripper(MultiFunction):
    """Terminal stripper."""

    def __init__(self):
        """Initialise."""
        super().__init__()
        self.mapping = {}

    def argument(self, o):
        """Apply to argument."""
        o_new = Argument(strip_function_space(o.ufl_function_space()), o.number(), o.part())
        return self.mapping.setdefault(o, o_new)

    def coefficient(self, o):
        """Apply to coefficient."""
        o_new = Coefficient(strip_function_space(o.ufl_function_space()), o.count())
        return self.mapping.setdefault(o, o_new)

    def constant(self, o):
        """Apply to constant."""
        o_new = Constant(strip_domain(o.ufl_domain()), o.ufl_shape, o.count())
        return self.mapping.setdefault(o, o_new)

    expr = MultiFunction.reuse_if_untouched


def strip_terminal_data(o):
    """Return a new form where all terminals have been replaced by UFL-only equivalents.

    This function is useful for forms containing augmented UFL objects that
    hold references to large data structures. These objects are be extracted
    into the mapping allowing the form to be cached without leaking memory.

    Args:
        o: The object to be stripped. This must either be a Form or Integral

    Returns:
        A 2-tuple containing an equivalent UFL-only object and a mapping
        allowing the original form to be reconstructed using replace_terminal_data

    """
    # We need to keep track of two maps because integrals store references to the
    # domain and ``replace`` expects only a mapping containing ``Expr`` objects.
    if isinstance(o, Form):
        integrals = []
        expr_map = {}
        domain_map = {}
        for integral in o.integrals():
            itg, (emap, dmap) = strip_terminal_data(integral)
            integrals.append(itg)
            expr_map.update(emap)
            domain_map.update(dmap)
        return Form(integrals), (expr_map, domain_map)
    elif isinstance(o, Integral):
        handler = TerminalStripper()
        integrand = map_expr_dag(handler, o.integrand())
        domain = strip_domain(o.ufl_domain())
        # invert the mapping so it can be passed straight into replace_terminal_data
        expr_map = {v: k for k, v in handler.mapping.items()}
        domain_map = {domain: o.ufl_domain()}
        return o.reconstruct(integrand, domain=domain), (expr_map, domain_map)
    else:
        raise ValueError("Only Form or Integral inputs expected")


def replace_terminal_data(o, mapping):
    """Return a new form where the terminals have been replaced using the provided mapping.

    Args:
        o: The object to have its terminals replaced. This must either be a Form or Integral.
        mapping: A mapping suitable for reconstructing the form such as the one
            returned by strip_terminal_data.

    Returns:
        The new form.
    """
    if isinstance(o, Form):
        return Form([replace_terminal_data(itg, mapping) for itg in o.integrals()])
    elif isinstance(o, Integral):
        expr_map, domain_map = mapping
        integrand = replace(o.integrand(), expr_map)
        return o.reconstruct(integrand, domain=domain_map[o.ufl_domain()])
    else:
        raise ValueError("Only Form or Integral inputs expected")


def strip_function_space(function_space):
    """Return a new function space with all non-UFL information removed."""
    if isinstance(function_space, FunctionSpace):
        return FunctionSpace(
            strip_domain(function_space.ufl_domain()), function_space.ufl_element()
        )
    elif isinstance(function_space, TensorProductFunctionSpace):
        subspaces = [strip_function_space(sub) for sub in function_space.ufl_sub_spaces()]
        return TensorProductFunctionSpace(*subspaces)
    elif isinstance(function_space, MixedFunctionSpace):
        subspaces = [strip_function_space(sub) for sub in function_space.ufl_sub_spaces()]
        return MixedFunctionSpace(*subspaces)
    else:
        raise NotImplementedError(f"{type(function_space)} cannot be stripped")


def strip_domain(domain):
    """Return a new domain with all non-UFL information removed."""
    if isinstance(domain, Mesh):
        return Mesh(domain.ufl_coordinate_element(), domain.ufl_id())
    elif isinstance(domain, MeshView):
        return MeshView(
            strip_domain(domain.ufl_mesh()), domain.topological_dimension(), domain.ufl_id()
        )
    else:
        raise NotImplementedError(f"{type(domain)} cannot be stripped")



"""This module defines special types for representing mapping of expressions to expressions."""
# Copyright (C) 2014 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

from ufl.argument import Coargument
from ufl.coefficient import Cofunction
from ufl.core.expr import Expr
from ufl.core.operator import Operator
from ufl.core.ufl_type import ufl_type

# --- Non-tensor types ---


@ufl_type(num_ops="varying")
class ExprList(Operator):
    """List of Expr objects. For internal use, never to be created by end users."""

    __slots__ = ()

    def __init__(self, *operands):
        """Initialise."""
        Operator.__init__(self, operands)
        # Enable Cofunction/Coargument for BaseForm differentiation
        if not all(isinstance(i, (Expr, Cofunction, Coargument)) for i in operands):
            raise ValueError("Expecting Expr, Cofunction or Coargument in ExprList.")

    def __getitem__(self, i):
        """Get an item."""
        return self.ufl_operands[i]

    def __len__(self):
        """Get the length."""
        return len(self.ufl_operands)

    def __iter__(self):
        """Return iterable."""
        return iter(self.ufl_operands)

    def __str__(self):
        """Format as a string."""
        return "ExprList(*(%s,))" % ", ".join(str(i) for i in self.ufl_operands)

    def __repr__(self):
        """Representation."""
        r = "ExprList(*%s)" % repr(self.ufl_operands)
        return r

    @property
    def ufl_shape(self):
        """Get the UFL shape."""
        raise ValueError("A non-tensor type has no ufl_shape.")

    @property
    def ufl_free_indices(self):
        """Get the UFL free indices."""
        raise ValueError("A non-tensor type has no ufl_free_indices.")

    def free_indices(self):
        """Get the free indices."""
        raise ValueError("A non-tensor type has no free_indices.")

    @property
    def ufl_index_dimensions(self):
        """Get the UFL index dimensions."""
        raise ValueError("A non-tensor type has no ufl_index_dimensions.")

    def index_dimensions(self):
        """Get the index dimensions."""
        raise ValueError("A non-tensor type has no index_dimensions.")


@ufl_type(num_ops="varying")
class ExprMapping(Operator):
    """Mapping of Expr objects. For internal use, never to be created by end users."""

    __slots__ = ()

    def __init__(self, *operands):
        """Initialise."""
        Operator.__init__(self, operands)
        if not all(isinstance(e, Expr) for e in operands):
            raise ValueError("Expecting Expr in ExprMapping.")

    def ufl_domains(self):
        """Get the UFL domains."""
        # Because this type can act like a terminal if it has no
        # operands, we need to override some recursive operations
        if self.ufl_operands:
            return Operator.ufl_domains()
        else:
            return []

    def __str__(self):
        """Format as a string."""
        return "ExprMapping(*%s)" % repr(self.ufl_operands)

    def __repr__(self):
        """Representation."""
        r = "ExprMapping(*%s)" % repr(self.ufl_operands)
        return r

    @property
    def ufl_shape(self):
        """Get the UFL shape."""
        raise ValueError("A non-tensor type has no ufl_shape.")

    @property
    def ufl_free_indices(self):
        """Get the UFL free indices."""
        raise ValueError("A non-tensor type has no ufl_free_indices.")

    def free_indices(self):
        """Get the free indices."""
        raise ValueError("A non-tensor type has no free_indices.")

    @property
    def ufl_index_dimensions(self):
        """Get the UFL index dimensions."""
        raise ValueError("A non-tensor type has no ufl_index_dimensions.")

    def index_dimensions(self):
        """Get the index dimensions."""
        raise ValueError("A non-tensor type has no index_dimensions.")



"""Algorithm for replacing gradients in an expression."""

# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

from ufl.algorithms.apply_function_pullbacks import apply_function_pullbacks
from ufl.algorithms.apply_geometry_lowering import apply_geometry_lowering
from ufl.checks import is_cellwise_constant
from ufl.classes import Grad, JacobianInverse, ReferenceGrad, ReferenceValue, Restricted
from ufl.core.multiindex import indices
from ufl.corealg.map_dag import map_expr_dag
from ufl.corealg.multifunction import MultiFunction
from ufl.domain import extract_unique_domain
from ufl.tensors import as_tensor

"""
# Some notes:
# Below, let v_i mean physical coordinate of vertex i and V_i mean the
reference cell coordinate of the same vertex.


# Add a type for CellVertices? Note that vertices must be computed in
linear cell cases! triangle_vertices[i,j] = component j of vertex i,
following ufc numbering conventions


# DONE Add a type for CellEdgeLengths? Note that these are only easy to
define in the linear cell case!
triangle_edge_lengths    = [v1v2, v0v2, v0v1] # shape (3,)
tetrahedron_edge_lengths = [v0v1, v0v2, v0v3, v1v2, v1v3, v2v3] # shape (6,)


# DONE Here's how to compute edge lengths from the Jacobian:
J =[ [dx0/dX0, dx0/dX1],
     [dx1/dX0, dx1/dX1] ]
# First compute the edge vector, which is constant for each edge: the
vector from one vertex to the other
reference_edge_vector_0 = V2 - V1 # Example! Add a type ReferenceEdgeVectors?
# Then apply J to it and take the length of the resulting vector, this is generic for affine cells
edge_length_i = || dot(J, reference_edge_vector_i) ||

e2 = || J[:,0] . < 1, 0> || = || J[:,0] || = || dx/dX0 || = edge length of edge 2 (v0-v1)
e1 = || J[:,1] . < 0, 1> || = || J[:,1] || = || dx/dX1 || = edge length of edge 1 (v0-v2)
e0 = || J[:,:] . <-1, 1> || = || < J[0,1]-J[0,0], J[1,1]-J[1,0] > || = || dx/dX <-1,1> ||
    = edge length of edge 0 (v1-v2)

trev = triangle_reference_edge_vector
evec0 = J00 * trev[edge][0] + J01 * trev[edge][1]  =  J*trev[edge]
evec1 = J10 * trev[edge][0] + J11 * trev[edge][1]
elen[edge] = sqrt(evec0*evec0 + evec1*evec1)  = sqrt((J*trev[edge])**2)

trev = triangle_reference_edge_vector
evec0 = J00 * trev[edge][0] + J01 * trev[edge][1]  =  J*trev
evec1 = J10 * trev[edge][0] + J11 * trev[edge][1]
evec2 = J20 * trev[edge][0] + J21 * trev[edge][1] # Manifold: triangle in 3D
elen[edge] = sqrt(evec0*evec0 + evec1*evec1 + evec2*evec2)  = sqrt((J*trev[edge])**2)

trev = tetrahedron_reference_edge_vector
evec0 = sum(J[0,k] * trev[edge][k] for k in range(3))
evec1 = sum(J[1,k] * trev[edge][k] for k in range(3))
evec2 = sum(J[2,k] * trev[edge][k] for k in range(3))
elen[edge] = sqrt(evec0*evec0 + evec1*evec1 + evec2*evec2)  = sqrt((J*trev[edge])**2)


# DONE Here's how to compute min/max facet edge length:
triangle:
  r = facetarea
tetrahedron:
  min(elen[edge] for edge in range(6))
or
  min( min(elen[0], min(elen[1], elen[2])), min(elen[3], min(elen[4], elen[5])) )
or
  min1 = min_value(elen[0], min_value(elen[1], elen[2]))
  min2 = min_value(elen[3], min_value(elen[4], elen[5]))
  r = min_value(min1, min2)
(want proper Min/Max types for this!)


# DONE Here's how to compute circumradius for an interval:
circumradius_interval = cellvolume / 2


# DONE Here's how to compute circumradius for a triangle:
e0 = elen[0]
e1 = elen[1]
e2 = elen[2]
circumradius_triangle = (e0*e1*e2) / (4*cellvolume)


# DONE Here's how to compute circumradius for a tetrahedron:
# v1v2 = edge length between vertex 1 and 2
# la,lb,lc = lengths of the sides of an intermediate triangle
la = v1v2 * v0v3
lb = v0v2 * v1v3
lc = v0v1 * v2v3
# p = perimeter
p = (la + lb + lc)
# s = semiperimeter
s = p / 2
# area of intermediate triangle with Herons formula
tmp_area = sqrt(s * (s - la) * (s - lb) * (s - lc))
circumradius_tetrahedron = tmp_area / (6*cellvolume)

"""


class ChangeToReferenceGrad(MultiFunction):
    """Change to reference grad."""

    def __init__(self):
        """Initalise."""
        MultiFunction.__init__(self)

    expr = MultiFunction.reuse_if_untouched

    def terminal(self, o):
        """Apply to terminal."""
        return o

    def grad(self, o):
        """Apply to grad."""
        # Peel off the Grads and count them, and get restriction if
        # it's between the grad and the terminal
        ngrads = 0
        restricted = ""
        rv = False
        while not o._ufl_is_terminal_:
            if isinstance(o, Grad):
                (o,) = o.ufl_operands
                ngrads += 1
            elif isinstance(o, Restricted):
                restricted = o.side()
                (o,) = o.ufl_operands
            elif isinstance(o, ReferenceValue):
                rv = True
                (o,) = o.ufl_operands
            else:
                raise ValueError(f"Invalid type {o._ufl_class_.__name__}")
        f = o
        if rv:
            f = ReferenceValue(f)

        # Get domain and create Jacobian inverse object
        domain = extract_unique_domain(o)
        Jinv = JacobianInverse(domain)

        if is_cellwise_constant(Jinv):
            # Optimise slightly by turning Grad(Grad(...)) into
            # J^(-T)J^(-T)RefGrad(RefGrad(...))
            # rather than J^(-T)RefGrad(J^(-T)RefGrad(...))

            # Create some new indices
            ii = indices(len(f.ufl_shape))  # Indices to get to the scalar component of f
            jj = indices(
                ngrads
            )  # Indices to sum over the local gradient axes with the inverse Jacobian
            kk = indices(ngrads)  # Indices for the leftover inverse Jacobian axes

            # Preserve restricted property
            if restricted:
                Jinv = Jinv(restricted)
                f = f(restricted)

            # Apply the same number of ReferenceGrad without mappings
            lgrad = f
            for i in range(ngrads):
                lgrad = ReferenceGrad(lgrad)

            # Apply mappings with scalar indexing operations (assumes
            # ReferenceGrad(Jinv) is zero)
            jinv_lgrad_f = lgrad[ii + jj]
            for j, k in zip(jj, kk):
                jinv_lgrad_f = Jinv[j, k] * jinv_lgrad_f

            # Wrap back in tensor shape, derivative axes at the end
            jinv_lgrad_f = as_tensor(jinv_lgrad_f, ii + kk)

        else:
            # J^(-T)RefGrad(J^(-T)RefGrad(...))

            # Preserve restricted property
            if restricted:
                Jinv = Jinv(restricted)
                f = f(restricted)

            jinv_lgrad_f = f
            for foo in range(ngrads):
                ii = indices(
                    len(jinv_lgrad_f.ufl_shape)
                )  # Indices to get to the scalar component of f
                j, k = indices(2)

                lgrad = ReferenceGrad(jinv_lgrad_f)
                jinv_lgrad_f = Jinv[j, k] * lgrad[ii + (j,)]

                # Wrap back in tensor shape, derivative axes at the end
                jinv_lgrad_f = as_tensor(jinv_lgrad_f, ii + (k,))

        return jinv_lgrad_f

    def reference_grad(self, o):
        """Apply to reference_grad."""
        raise ValueError("Not expecting reference grad while applying change to reference grad.")

    def coefficient_derivative(self, o):
        """Apply to coefficient_derivative."""
        raise ValueError(
            "Coefficient derivatives should be expanded before applying change to reference grad."
        )


def change_to_reference_grad(e):
    """Change Grad objects in expression to products of JacobianInverse and ReferenceGrad.

    Assumes the expression is preprocessed or at least that derivatives
    have been expanded.

    Args:
        e: An Expr or Form.
    """
    mf = ChangeToReferenceGrad()
    return map_expr_dag(mf, e)


def change_integrand_geometry_representation(integrand, scale, integral_type):
    """Change integrand geometry to the right representations."""
    integrand = apply_function_pullbacks(integrand)

    integrand = change_to_reference_grad(integrand)

    integrand = integrand * scale

    if integral_type == "quadrature":
        physical_coordinates_known = True
    else:
        physical_coordinates_known = False
    integrand = apply_geometry_lowering(integrand, physical_coordinates_known)

    return integrand



#
# Author: Martin Sandve Alnes
# Date: 2008-12-22
#

from ufl import (
    Coefficient,
    Constant,
    FacetNormal,
    FunctionSpace,
    Identity,
    Mesh,
    SpatialCoordinate,
    TestFunction,
    TrialFunction,
    derivative,
    det,
    diff,
    dot,
    ds,
    dx,
    exp,
    grad,
    inner,
    inv,
    tetrahedron,
    tr,
    variable,
)
from ufl.finiteelement import FiniteElement

# Modified by Garth N. Wells, 2009
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1

# Cell and its properties
cell = tetrahedron
domain = Mesh(FiniteElement("Lagrange", cell, 1, (3,), identity_pullback, H1))
d = 3
N = FacetNormal(domain)
x = SpatialCoordinate(domain)

# Elements
u_element = FiniteElement("Lagrange", cell, 2, (3,), identity_pullback, H1)
p_element = FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1)
A_element = FiniteElement("Lagrange", cell, 1, (3, 3), identity_pullback, H1)

# Spaces
u_space = FunctionSpace(domain, u_element)
p_space = FunctionSpace(domain, p_element)
A_space = FunctionSpace(domain, A_element)

# Test and trial functions
v = TestFunction(u_space)
w = TrialFunction(u_space)

# Displacement at current and two previous timesteps
u = Coefficient(u_space)
up = Coefficient(u_space)
upp = Coefficient(u_space)

# Time parameters
dt = Constant(domain)

# Fiber field
A = Coefficient(A_space)

# External forces
T = Coefficient(u_space)
p0 = Coefficient(p_space)

# Material parameters FIXME
rho = Constant(domain)
K = Constant(domain)
c00 = Constant(domain)
c11 = Constant(domain)
c22 = Constant(domain)

# Deformation gradient
Id = Identity(d)
F = Id + grad(u)
F = variable(F)
Finv = inv(F)
J = det(F)

# Left Cauchy-Green deformation tensor
B = F * F.T
I1_B = tr(B)
I2_B = (I1_B**2 - tr(B * B)) / 2
I3_B = J**2

# Right Cauchy-Green deformation tensor
C = F.T * F
I1_C = tr(C)
I2_C = (I1_C**2 - tr(C * C)) / 2
I3_C = J**2

# Green strain tensor
E = (C - Id) / 2

# Mapping of strain in fiber directions
Ef = A * E * A.T

# Strain energy function W(Q(Ef))
Q = (
    c00 * Ef[0, 0] ** 2 + c11 * Ef[1, 1] ** 2 + c22 * Ef[2, 2] ** 2
)  # FIXME: insert some simple law here
W = (K / 2) * (exp(Q) - 1)  # + p stuff

# First Piola-Kirchoff stress tensor
P = diff(W, F)

# Acceleration term discretized with finite differences
k = dt / rho
acc = u - 2 * up + upp

# Residual equation # FIXME: Can contain errors, not tested!
a_F = (
    inner(acc, v) * dx
    + k * inner(P, grad(v)) * dx
    - k * dot(J * Finv * T, v) * ds(0)
    - k * dot(J * Finv * p0 * N, v) * ds(1)
)

# Jacobi matrix of residual equation
a_J = derivative(a_F, u, w)

# Export forms
forms = [a_F, a_J]



"""Sorting.

This module contains a sorting rule for expr objects that
is more robust w.r.t. argument numbering than using repr.
"""
# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Anders Logg, 2009-2010.
# Modified by Johan Hake, 2010.

from functools import cmp_to_key

from ufl.argument import Argument
from ufl.coefficient import Coefficient
from ufl.core.multiindex import FixedIndex, MultiIndex
from ufl.variable import Label


def _cmp_multi_index(a, b):
    """Cmp multi index."""
    # Careful not to depend on Index.count() here!
    # This is placed first because it is most frequent.
    # Make decision based on the first index pair possible
    for i, j in zip(a._indices, b._indices):
        fix1 = isinstance(i, FixedIndex)
        fix2 = isinstance(j, FixedIndex)
        if fix1 and fix2:
            # Both are FixedIndex, sort by value
            x, y = i._value, j._value
            if x < y:
                return -1
            elif x > y:
                return 1
            else:
                # Same value, no decision
                continue
        elif fix1:
            # Sort fixed before free
            return -1
        elif fix2:
            # Sort fixed before free
            return 1
        else:
            # Both are Index, no decision, do not depend on count!
            pass
    # Failed to make a decision, return 0 by default
    # (this does not mean equality, it could be e.g.
    # [i,0] vs [j,0] because the counts of i,j cannot be used)
    return 0


def _cmp_label(a, b):
    """Cmp label."""
    # Don't compare counts! Causes circular problems when renumbering to get a canonical form.
    # Therefore, even though a and b are not equal in general (__eq__ won't be True),
    # but for this sorting they are considered equal and we return 0.
    return 0


def _cmp_coefficient(a, b):
    """Cmp coefficient."""
    # It's ok to compare relative counts for Coefficients,
    # since their ordering is a property of the form
    x, y = a._count, b._count
    if x < y:
        return -1
    elif x > y:
        return 1
    else:
        return 0


def _cmp_argument(a, b):
    """Cmp argument."""
    # It's ok to compare relative number and part for Arguments,
    # since their ordering is a property of the form
    x = (a._number, a._part)
    y = (b._number, b._part)
    if x < y:
        return -1
    elif x > y:
        return 1
    else:
        return 0


def _cmp_terminal_by_repr(a, b):
    """Cmp terminal by repr."""
    # The cost of repr on a terminal is fairly small, and bounded
    x = repr(a)
    y = repr(b)
    return -1 if x < y else (0 if x == y else 1)


# Hack up a MultiFunction-like type dispatch for terminal comparisons
_terminal_cmps = {}
_terminal_cmps[MultiIndex._ufl_typecode_] = _cmp_multi_index
_terminal_cmps[Argument._ufl_typecode_] = _cmp_argument
_terminal_cmps[Coefficient._ufl_typecode_] = _cmp_coefficient
_terminal_cmps[Label._ufl_typecode_] = _cmp_label


def cmp_expr(a, b):
    """Replacement for cmp(a, b), removed in Python 3, for Expr objects."""
    # Modelled after pre_traversal to avoid recursion:
    left = [(a, b)]
    while left:
        a, b = left.pop()

        # First sort quickly by type code
        x, y = a._ufl_typecode_, b._ufl_typecode_
        if x != y:
            return -1 if x < y else 1

        # Now we know that the type is the same, check further based
        # on type specific properties.
        if a._ufl_is_terminal_:
            if x in _terminal_cmps:
                c = _terminal_cmps[x](a, b)
            else:
                c = _cmp_terminal_by_repr(a, b)

            if c:
                return c
        else:
            # If the hash is the same, assume equal for the purpose of
            # sorting.  This introduces a minor chance of
            # nondeterministic behaviour, just as with MultiIndex.
            # Although collected statistics for complicated forms
            # suggest that the hash function is pretty good so there
            # shouldn't be collisions.
            # if hash(a) == hash(b): # FIXME: Test this for performance improvement.
            #    return 0

            # Delve into subtrees
            aops = a.ufl_operands
            bops = b.ufl_operands

            # Sort by children in natural order
            for r, s in zip(aops, bops):
                # Skip subtree if objects are the same
                if r is s:
                    continue
                # Append subtree for further inspection
                left.append((r, s))

            # All children compare as equal, a and b must be
            # equal. Except for...  A few types, notably ExprList and
            # ExprMapping, can have a different number of operands.
            # Sort by the length if it's different. Doing this after
            # sorting by children because these types are rare so we
            # try to avoid the cost of this check for most nodes.
            x, y = len(aops), len(bops)
            if x != y:
                return -1 if x < y else 1

    # Equal if we get out of the above loop!
    return 0


def sorted_expr(sequence):
    """Return a canonically sorted list of Expr objects in sequence."""
    return sorted(sequence, key=cmp_to_key(cmp_expr))


def sorted_expr_sum(seq):
    """Sorted expr sum."""
    seq2 = sorted(seq, key=cmp_to_key(cmp_expr))
    s = seq2[0]
    for e in seq2[1:]:
        s = s + e
    return s



"""Apply restrictions.

This module contains the apply_restrictions algorithm which propagates
restrictions in a form towards the terminals.
"""

# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

from ufl.algorithms.map_integrands import map_integrand_dags
from ufl.classes import Restricted
from ufl.corealg.map_dag import map_expr_dag
from ufl.corealg.multifunction import MultiFunction
from ufl.domain import extract_unique_domain
from ufl.measure import integral_type_to_measure_name
from ufl.sobolevspace import H1


class RestrictionPropagator(MultiFunction):
    """Restriction propagator."""

    def __init__(self, side=None):
        """Initialise."""
        MultiFunction.__init__(self)
        self.current_restriction = side
        self.default_restriction = "+"
        # Caches for propagating the restriction with map_expr_dag
        self.vcaches = {"+": {}, "-": {}}
        self.rcaches = {"+": {}, "-": {}}
        if self.current_restriction is None:
            self._rp = {"+": RestrictionPropagator("+"), "-": RestrictionPropagator("-")}

    def restricted(self, o):
        """When hitting a restricted quantity, visit child with a separate restriction algorithm."""
        # Assure that we have only two levels here, inside or outside
        # the Restricted node
        if self.current_restriction is not None:
            raise ValueError("Cannot restrict an expression twice.")
        # Configure a propagator for this side and apply to subtree
        side = o.side()
        return map_expr_dag(
            self._rp[side], o.ufl_operands[0], vcache=self.vcaches[side], rcache=self.rcaches[side]
        )

    # --- Reusable rules

    def _ignore_restriction(self, o):
        """Ignore current restriction.

        Quantity is independent of side also from a computational point
        of view.
        """
        return o

    def _require_restriction(self, o):
        """Restrict a discontinuous quantity to current side, require a side to be set."""
        if self.current_restriction is None:
            raise ValueError(f"Discontinuous type {o._ufl_class_.__name__} must be restricted.")
        return o(self.current_restriction)

    def _default_restricted(self, o):
        """Restrict a continuous quantity to default side if no current restriction is set."""
        r = self.current_restriction
        if r is None:
            r = self.default_restriction
        return o(r)

    def _opposite(self, o):
        """Restrict a quantity to default side.

        If the current restriction is different swap the sign, require a side to be set.
        """
        if self.current_restriction is None:
            raise ValueError(f"Discontinuous type {o._ufl_class_.__name__} must be restricted.")
        elif self.current_restriction == self.default_restriction:
            return o(self.default_restriction)
        else:
            return -o(self.default_restriction)

    def _missing_rule(self, o):
        """Raise an error."""
        raise ValueError(f"Missing rule for {o._ufl_class_.__name__}")

    # --- Rules for operators

    # Default: Operators should reconstruct only if subtrees are not touched
    operator = MultiFunction.reuse_if_untouched

    # Assuming apply_derivatives has been called,
    # propagating Grad inside the Restricted nodes.
    # Considering all grads to be discontinuous, may
    # want something else for facet functions in future.
    grad = _require_restriction

    def variable(self, o, op, label):
        """Strip variable."""
        return op

    def reference_value(self, o):
        """Reference value of something follows same restriction rule as the underlying object."""
        (f,) = o.ufl_operands
        assert f._ufl_is_terminal_
        g = self(f)
        if isinstance(g, Restricted):
            side = g.side()
            return o(side)
        else:
            return o

    # --- Rules for terminals

    # Require handlers to be specified for all terminals
    terminal = _missing_rule

    multi_index = _ignore_restriction
    label = _ignore_restriction

    # Default: Literals should ignore restriction
    constant_value = _ignore_restriction
    constant = _ignore_restriction

    # Even arguments with continuous elements such as Lagrange must be
    # restricted to associate with the right part of the element
    # matrix
    argument = _require_restriction

    # Defaults for geometric quantities
    geometric_cell_quantity = _require_restriction
    geometric_facet_quantity = _require_restriction

    # Only a few geometric quantities are independent on the restriction:
    facet_coordinate = _ignore_restriction
    quadrature_weight = _ignore_restriction

    # Assuming homogeoneous mesh
    reference_cell_volume = _ignore_restriction
    reference_facet_volume = _ignore_restriction

    def coefficient(self, o):
        """Restrict a coefficient.

        Allow coefficients to be unrestricted (apply default if so) if
        the values are fully continuous across the facet.
        """
        if o.ufl_element() in H1:
            # If the coefficient _value_ is _fully_ continuous
            # It must still be computed from one of the sides, we just don't care which
            return self._default_restricted(o)
        else:
            return self._require_restriction(o)

    def facet_normal(self, o):
        """Restrict a facet_normal."""
        D = extract_unique_domain(o)
        e = D.ufl_coordinate_element()
        gd = D.geometric_dimension()
        td = D.topological_dimension()

        if e.embedded_superdegree <= 1 and e in H1 and gd == td:
            # For meshes with a continuous linear non-manifold
            # coordinate field, the facet normal from side - points in
            # the opposite direction of the one from side +.  We must
            # still require a side to be chosen by the user but
            # rewrite n- -> n+.  This is an optimization, possibly
            # premature, however it's more difficult to do at a later
            # stage.
            return self._opposite(o)
        else:
            # For other meshes, we require a side to be
            # chosen by the user and respect that
            return self._require_restriction(o)


def apply_restrictions(expression):
    """Propagate restriction nodes to wrap differential terminals directly."""
    integral_types = [
        k for k in integral_type_to_measure_name.keys() if k.startswith("interior_facet")
    ]
    rules = RestrictionPropagator()
    return map_integrand_dags(rules, expression, only_integral_type=integral_types)


class DefaultRestrictionApplier(MultiFunction):
    """Default restriction applier."""

    def __init__(self, side=None):
        """Initialise."""
        MultiFunction.__init__(self)
        self.current_restriction = side
        self.default_restriction = "+"
        if self.current_restriction is None:
            self._rp = {"+": DefaultRestrictionApplier("+"), "-": DefaultRestrictionApplier("-")}

    def terminal(self, o):
        """Apply to terminal."""
        # Most terminals are unchanged
        return o

    # Default: Operators should reconstruct only if subtrees are not touched
    operator = MultiFunction.reuse_if_untouched

    def restricted(self, o):
        """Apply to restricted."""
        # Don't restrict twice
        return o

    def derivative(self, o):
        """Apply to derivative."""
        # I don't think it's safe to just apply default restriction
        # to the argument of any derivative, i.e. grad(cg1_function)
        # is not continuous across cells even if cg1_function is.
        return o

    def _default_restricted(self, o):
        """Restrict a continuous quantity to default side if no current restriction is set."""
        r = self.current_restriction
        if r is None:
            r = self.default_restriction
        return o(r)

    # These are the same from either side but to compute them
    # cell (or facet) data from one side must be selected:
    spatial_coordinate = _default_restricted
    # Depends on cell only to get to the facet:
    facet_jacobian = _default_restricted
    facet_jacobian_determinant = _default_restricted
    facet_jacobian_inverse = _default_restricted
    # facet_tangents = _default_restricted
    # facet_midpoint = _default_restricted
    facet_area = _default_restricted
    # facet_diameter = _default_restricted
    min_facet_edge_length = _default_restricted
    max_facet_edge_length = _default_restricted
    facet_origin = _default_restricted  # FIXME: Is this valid for quads?


def apply_default_restrictions(expression):
    """Some terminals can be restricted from either side.

    This applies a default restriction to such terminals if unrestricted.
    """
    integral_types = [
        k for k in integral_type_to_measure_name.keys() if k.startswith("interior_facet")
    ]
    rules = DefaultRestrictionApplier()
    return map_integrand_dags(rules, expression, only_integral_type=integral_types)



"""The Equation class, used to express equations like a == L."""
# Copyright (C) 2012-2016 Anders Logg and Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

# Export list for ufl.classes
__all_classes__ = ["Equation"]


class Equation(object):
    """Equation.

    This class is used to represent equations expressed by the "=="
    operator. Examples include a == L and F == 0 where a, L and F are
    Form objects.
    """

    def __init__(self, lhs, rhs):
        """Create equation lhs == rhs."""
        self.lhs = lhs
        self.rhs = rhs

    def __bool__(self):
        """Evaluate bool(lhs_form == rhs_form).

        This will not trigger when setting 'equation = a == L',
        but when e.g. running 'if equation:'.
        """
        # NB!: pep8 will say you should use isinstance here, but we do
        #      actually want to compare the exact types in this case.
        # Not equal if types are not identical (i.e. not accepting
        # subclasses)
        if type(self.lhs) != type(self.rhs):  # noqa: E721
            return False
        # Try to delegate to equals function
        if hasattr(self.lhs, "equals"):
            return self.lhs.equals(self.rhs)
        elif hasattr(self.rhs, "equals"):
            return self.rhs.equals(self.lhs)
        else:
            raise ValueError("Either lhs or rhs of Equation must implement self.equals(other).")

    __nonzero__ = __bool__

    def __eq__(self, other):
        """Compare two equations by comparing lhs and rhs."""
        return isinstance(other, Equation) and self.lhs == other.lhs and self.rhs == other.rhs

    def __hash__(self):
        """Hash."""
        return hash((hash(self.lhs), hash(self.rhs)))

    def __repr__(self):
        """Representation."""
        return f"Equation({self.lhs!r}, {self.rhs!r})"



# Copyright (C) 2008-2010 Marie Rognes
#
# This file is part of UFL.
#
# UFL is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UFL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with UFL. If not, see <http://www.gnu.org/licenses/>.
#
# First added:  2008-10-03
# Last changed: 2011-07-22
from ufl import (
    FunctionSpace,
    Mesh,
    TestFunctions,
    TrialFunctions,
    as_vector,
    div,
    dot,
    dx,
    inner,
    skew,
    tetrahedron,
    tr,
)
from ufl.finiteelement import FiniteElement, MixedElement
from ufl.pullback import contravariant_piola, identity_pullback
from ufl.sobolevspace import H1, L2, HDiv


def skw(tau):
    """Define vectorized skew operator"""
    sk = 2 * skew(tau)
    return as_vector((sk[0, 1], sk[0, 2], sk[1, 2]))


cell = tetrahedron
n = 3

# Finite element exterior calculus syntax
r = 1
S = FiniteElement("vector BDM", cell, r, (3, 3), contravariant_piola, HDiv)
V = FiniteElement("Discontinuous Lagrange", cell, r - 1, (3,), identity_pullback, L2)
Q = FiniteElement("Discontinuous Lagrange", cell, r - 1, (3,), identity_pullback, L2)

W = MixedElement([S, V, Q])

domain = Mesh(FiniteElement("Lagrange", cell, 1, (3,), identity_pullback, H1))
space = FunctionSpace(domain, W)

(sigma, u, gamma) = TrialFunctions(space)
(tau, v, eta) = TestFunctions(space)

a = (
    inner(sigma, tau)
    - tr(sigma) * tr(tau)
    + dot(div(tau), u)
    - dot(div(sigma), v)
    + inner(skw(tau), gamma)
    + inner(skw(sigma), eta)
) * dx



#
# Harmonic map demo using separate coefficients x and y.
# Author: Martin Alnes
# Date: 2009-04-09
#
from ufl import Coefficient, FunctionSpace, Mesh, derivative, dot, dx, grad, inner, triangle
from ufl.finiteelement import FiniteElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1

cell = triangle
X = FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1)
Y = FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1)
domain = Mesh(FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1))
X_space = FunctionSpace(domain, X)
Y_space = FunctionSpace(domain, Y)

x = Coefficient(X_space)
y = Coefficient(Y_space)

L = inner(grad(x), grad(x)) * dx + dot(x, x) * y * dx

F = derivative(L, (x, y))
J = derivative(F, (x, y))

forms = [L, F, J]



"""Algorithms for building canonical data structure for integrals over subdomains."""

# Copyright (C) 2009-2016 Anders Logg and Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

import numbers
import typing
from collections import defaultdict

import ufl
from ufl.algorithms.coordinate_derivative_helpers import (
    attach_coordinate_derivatives,
    strip_coordinate_derivatives,
)
from ufl.algorithms.renumbering import renumber_indices
from ufl.form import Form
from ufl.integral import Integral
from ufl.protocols import id_or_none
from ufl.sorting import cmp_expr, sorted_expr
from ufl.utils.sorting import canonicalize_metadata, sorted_by_key


class IntegralData(object):
    """Utility class.

    This class has members (domain, integral_type, subdomain_id,
    integrals, metadata), where metadata is an empty dictionary that may
    be used for associating metadata with each object.
    """

    __slots__ = (
        "domain",
        "integral_type",
        "subdomain_id",
        "integrals",
        "metadata",
        "integral_coefficients",
        "enabled_coefficients",
    )

    def __init__(self, domain, integral_type, subdomain_id, integrals, metadata):
        """Initialise."""
        if 1 != len(set(itg.ufl_domain() for itg in integrals)):
            raise ValueError("Multiple domains mismatch in integral data.")
        if not all(integral_type == itg.integral_type() for itg in integrals):
            raise ValueError("Integral type mismatch in integral data.")
        if not all(subdomain_id == itg.subdomain_id() for itg in integrals):
            raise ValueError("Subdomain id mismatch in integral data.")

        self.domain = domain
        self.integral_type = integral_type
        self.subdomain_id = subdomain_id

        self.integrals = integrals

        # This is populated in preprocess using data not available at
        # this stage:
        self.integral_coefficients = None
        self.enabled_coefficients = None

        # TODO: I think we can get rid of this with some refactoring
        # in ffc:
        self.metadata = metadata

    def __lt__(self, other):
        """Check if self is less than other."""
        # To preserve behaviour of extract_integral_data:
        return (self.integral_type, self.subdomain_id, self.integrals, self.metadata) < (
            other.integral_type,
            other.subdomain_id,
            other.integrals,
            other.metadata,
        )

    def __eq__(self, other):
        """Check for equality."""
        # Currently only used for tests:
        return (
            self.integral_type == other.integral_type
            and self.subdomain_id == other.subdomain_id
            and self.integrals == other.integrals
            and self.metadata == other.metadata
        )

    def __str__(self):
        """Format as a string."""
        s = f"IntegralData over domain({self.integral_type}, {self.subdomain_id})"
        s += " with integrals:\n"
        s += "\n\n".join(map(str, self.integrals))
        s += "\nand metadata:\n{metadata}"
        return s


class ExprTupleKey(object):
    """Tuple comparison helper."""

    __slots__ = ("x",)

    def __init__(self, x):
        """Initialise."""
        self.x = x

    def __lt__(self, other):
        """Check if self is less than other."""
        # Comparing expression first
        c = cmp_expr(self.x[0], other.x[0])
        if c < 0:
            return True
        elif c > 0:
            return False
        else:
            # Comparing form compiler data
            mds = canonicalize_metadata(self.x[1])
            mdo = canonicalize_metadata(other.x[1])
            return mds < mdo


def group_integrals_by_domain_and_type(integrals, domains):
    """Group integrals by domain and type.

    Args:
        integrals: list of Integral objects
        domains: list of AbstractDomain objects from the parent Form

    Returns:
        Dictionary mapping (domain, integral_type) to list(Integral)
    """
    integrals_by_domain_and_type = defaultdict(list)
    for itg in integrals:
        if itg.ufl_domain() is None:
            raise ValueError("Integral has no domain.")
        key = (itg.ufl_domain(), itg.integral_type())

        # Append integral to list of integrals with shared key
        integrals_by_domain_and_type[key].append(itg)

    return integrals_by_domain_and_type


def integral_subdomain_ids(integral):
    """Get a tuple of integer subdomains or a valid string subdomain from integral."""
    did = integral.subdomain_id()
    if isinstance(did, numbers.Integral):
        return (did,)
    elif isinstance(did, tuple):
        if not all(isinstance(d, numbers.Integral) for d in did):
            raise ValueError("Expecting only integer subdomains in tuple.")
        return did
    elif did in ("everywhere", "otherwise"):
        # TODO: Define list of valid strings somewhere more central
        return did
    else:
        raise ValueError(f"Invalid domain id {did}.")


def rearrange_integrals_by_single_subdomains(
    integrals: typing.List[Integral], do_append_everywhere_integrals: bool
) -> typing.Dict[int, typing.List[Integral]]:
    """Rearrange integrals over multiple subdomains to single subdomain integrals.

    Args:
        integrals: List of integrals
        do_append_everywhere_integrals: Boolean indicating if integrals
        defined on the whole domain should
            just be restricted to the set of input subdomain ids.

    Returns:
        The integrals reconstructed with single subdomain_id
    """
    # Split integrals into lists of everywhere and subdomain integrals
    everywhere_integrals = []
    subdomain_integrals = []
    for itg in integrals:
        dids = integral_subdomain_ids(itg)
        if dids == "otherwise":
            raise ValueError("'otherwise' integrals should never occur before preprocessing.")
        elif dids == "everywhere":
            everywhere_integrals.append(itg)
        else:
            subdomain_integrals.append((dids, itg))

    # Fill single_subdomain_integrals with lists of integrals from
    # subdomain_integrals, but split and restricted to single
    # subdomain ids
    single_subdomain_integrals = defaultdict(list)
    for dids, itg in subdomain_integrals:
        # Region or single subdomain id
        for did in dids:
            # Restrict integral to this subdomain!
            single_subdomain_integrals[did].append(itg.reconstruct(subdomain_id=did))

    # Add everywhere integrals to each single subdomain id integral
    # list
    otherwise_integrals = []
    for ev_itg in everywhere_integrals:
        # Restrict everywhere integral to 'otherwise'
        otherwise_integrals.append(ev_itg.reconstruct(subdomain_id="otherwise"))

        # Restrict everywhere integral to each subdomain
        # and append to each integral list
        if do_append_everywhere_integrals:
            for subdomain_id in sorted(single_subdomain_integrals.keys()):
                single_subdomain_integrals[subdomain_id].append(
                    ev_itg.reconstruct(subdomain_id=subdomain_id)
                )

    if otherwise_integrals:
        single_subdomain_integrals["otherwise"] = otherwise_integrals

    return single_subdomain_integrals


def accumulate_integrands_with_same_metadata(integrals):
    """Accumulate integrands with the same metedata.

    Args:
        integrals: a list of integrals

    Returns:
        A list of the form [(integrand0, metadata0), (integrand1,
        metadata1), ...] where integrand0 < integrand1 by the canonical
        ufl expression ordering criteria.
    """
    # Group integrals by compiler data hash
    by_cdid = {}
    for itg in integrals:
        cd = itg.metadata()
        cdid = hash(canonicalize_metadata(cd))
        if cdid not in by_cdid:
            by_cdid[cdid] = ([], cd)
        by_cdid[cdid][0].append(itg)

    # Accumulate integrands separately for each compiler data object
    # id
    for cdid in by_cdid:
        integrals, cd = by_cdid[cdid]
        # Ensure canonical sorting of more than two integrands
        integrands = sorted_expr((itg.integrand() for itg in integrals))
        integrands_sum = sum(integrands[1:], integrands[0])
        by_cdid[cdid] = (integrands_sum, cd)

    # Sort integrands canonically by integrand first then compiler
    # data
    return sorted(by_cdid.values(), key=ExprTupleKey)


def build_integral_data(integrals):
    """Build integral data given a list of integrals.

    The integrals you pass in here must have been rearranged and
    gathered (removing the "everywhere" subdomain_id). To do this, you
    should call group_form_integrals.

    Args:
        integrals: An iterable of Integral objects.

    Returns:
        A tuple of IntegralData objects.
    """
    itgs = defaultdict(list)

    # --- Merge integral data that has the same integrals,
    for integral in integrals:
        integral_type = integral.integral_type()
        ufl_domain = integral.ufl_domain()
        subdomain_ids = integral.subdomain_id()
        if "everywhere" in subdomain_ids:
            raise ValueError(
                "'everywhere' not a valid subdomain id. "
                "Did you forget to call group_form_integrals?"
            )

        # Group for integral data (One integral data object for all
        # integrals with same domain, itype, (but possibly different metadata).
        itgs[(ufl_domain, integral_type, subdomain_ids)].append(integral)

    # Build list with canonical ordering, iteration over dicts
    # is not deterministic across python versions
    def keyfunc(item):
        (d, itype, sid), integrals = item
        sid_int = tuple(-1 if i == "otherwise" else i for i in sid)
        return (d._ufl_sort_key_(), itype, (type(sid).__name__,), sid_int)

    integral_datas = []
    for (d, itype, sid), integrals in sorted(itgs.items(), key=keyfunc):
        integral_datas.append(IntegralData(d, itype, sid, integrals, {}))
    return integral_datas


def group_form_integrals(form, domains, do_append_everywhere_integrals=True):
    """Group integrals by domain and type, performing canonical simplification.

    Args:
        form: the Form to group the integrals of.
        domains: an iterable of Domains.
        do_append_everywhere_integrals: Boolean indicating if integrals
        defined on the whole domain should
            just be restricted to the set of input subdomain ids.

    Returns:
        A new Form with gathered integrands.
    """
    # Group integrals by domain and type
    integrals_by_domain_and_type = group_integrals_by_domain_and_type(form.integrals(), domains)

    integrals = []
    for domain in domains:
        for integral_type in ufl.measure.integral_types():
            # Get integrals with this domain and type
            ddt_integrals = integrals_by_domain_and_type.get((domain, integral_type))
            if ddt_integrals is None:
                continue

            # Group integrals by subdomain id, after splitting e.g.
            #   f*dx((1,2)) + g*dx((2,3)) -> f*dx(1) + (f+g)*dx(2) + g*dx(3)
            # (note: before this call, 'everywhere' is a valid subdomain_id,
            # and after this call, 'otherwise' is a valid subdomain_id)
            single_subdomain_integrals = rearrange_integrals_by_single_subdomains(
                ddt_integrals, do_append_everywhere_integrals
            )

            for subdomain_id, ss_integrals in sorted_by_key(single_subdomain_integrals):
                # strip the coordinate derivatives from all integrals
                # this yields a list of the form [(coordinate derivative, integral), ...]
                stripped_integrals_and_coordderivs = strip_coordinate_derivatives(ss_integrals)

                # now group the integrals by the coordinate derivative
                def calc_hash(cd):
                    return sum(
                        sum(tuple_elem._ufl_compute_hash_() for tuple_elem in tuple_)
                        for tuple_ in cd
                    )

                coordderiv_integrals_dict = {}
                for integral, coordderiv in stripped_integrals_and_coordderivs:
                    coordderivhash = calc_hash(coordderiv)
                    if coordderivhash in coordderiv_integrals_dict:
                        coordderiv_integrals_dict[coordderivhash][1].append(integral)
                    else:
                        coordderiv_integrals_dict[coordderivhash] = (coordderiv, [integral])

                # cd_integrals_dict is now a dict of the form
                # { hash: (CoordinateDerivative, [integral, integral, ...]), ... }
                # we can now put the integrals back together and then afterwards
                # apply the CoordinateDerivative again

                for cdhash, samecd_integrals in sorted_by_key(coordderiv_integrals_dict):
                    # Accumulate integrands of integrals that share the
                    # same compiler data
                    integrands_and_cds = accumulate_integrands_with_same_metadata(
                        samecd_integrals[1]
                    )

                    for integrand, metadata in integrands_and_cds:
                        integral = Integral(
                            integrand, integral_type, domain, subdomain_id, metadata, None
                        )
                        integral = attach_coordinate_derivatives(integral, samecd_integrals[0])
                        integrals.append(integral)

    # Group integrals by common integrand
    # u.dx(0)*dx(1) + u.dx(0)*dx(2) -> u.dx(0)*dx((1,2))
    # to avoid duplicate kernels generated after geometry lowering
    unique_integrals = defaultdict(tuple)
    metadata_table = defaultdict(dict)
    for integral in integrals:
        integral_type = integral.integral_type()
        ufl_domain = integral.ufl_domain()
        metadata = integral.metadata()
        meta_hash = hash(canonicalize_metadata(metadata))
        subdomain_id = integral.subdomain_id()
        subdomain_data = id_or_none(integral.subdomain_data())
        integrand = renumber_indices(integral.integrand())
        unique_integrals[(integral_type, ufl_domain, meta_hash, integrand, subdomain_data)] += (
            subdomain_id,
        )
        metadata_table[(integral_type, ufl_domain, meta_hash, integrand, subdomain_data)] = metadata

    grouped_integrals = []
    for integral_data, subdomain_ids in unique_integrals.items():
        (integral_type, ufl_domain, metadata, integrand, subdomain_data) = integral_data
        integral = Integral(
            integrand,
            integral_type,
            ufl_domain,
            subdomain_ids,
            metadata_table[integral_data],
            subdomain_data,
        )
        grouped_integrals.append(integral)

    return Form(grouped_integrals)


def reconstruct_form_from_integral_data(integral_data):
    """Reconstruct a form from integral data."""
    integrals = []
    for ida in integral_data:
        integrals.extend(ida.integrals)
    return Form(integrals)



"""This module provides the compute_form_data function.

Form compilers will typically call compute_form_dataprior to code
generation to preprocess/simplify a raw input form given by a user.
"""
# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

from itertools import chain

from ufl.algorithms.analysis import extract_coefficients, extract_sub_elements, unique_tuple
from ufl.algorithms.apply_algebra_lowering import apply_algebra_lowering
from ufl.algorithms.apply_derivatives import apply_coordinate_derivatives, apply_derivatives

# These are the main symbolic processing steps:
from ufl.algorithms.apply_function_pullbacks import apply_function_pullbacks
from ufl.algorithms.apply_geometry_lowering import apply_geometry_lowering
from ufl.algorithms.apply_integral_scaling import apply_integral_scaling
from ufl.algorithms.apply_restrictions import apply_default_restrictions, apply_restrictions
from ufl.algorithms.check_arities import check_form_arity
from ufl.algorithms.comparison_checker import do_comparison_check

# See TODOs at the call sites of these below:
from ufl.algorithms.domain_analysis import (
    build_integral_data,
    group_form_integrals,
    reconstruct_form_from_integral_data,
)
from ufl.algorithms.estimate_degrees import estimate_total_polynomial_degree
from ufl.algorithms.formdata import FormData
from ufl.algorithms.formtransformations import compute_form_arities
from ufl.algorithms.remove_complex_nodes import remove_complex_nodes
from ufl.classes import Coefficient, Form, FunctionSpace, GeometricFacetQuantity
from ufl.corealg.traversal import traverse_unique_terminals
from ufl.domain import extract_unique_domain
from ufl.utils.sequences import max_degree


def _auto_select_degree(elements):
    """Automatically select degree for all elements of the form.

    This is be used in cases where the degree has not been specified by the user.
    This feature is used by DOLFIN to allow the specification of Expressions with
    undefined degrees.
    """
    # Use max degree of all elements, at least 1 (to work with
    # Lagrange elements)
    return max_degree({e.embedded_superdegree for e in elements} - {None} | {1})


def _compute_element_mapping(form):
    """Compute element mapping for element replacement."""
    # The element mapping is a slightly messy concept with two use
    # cases:
    # - Expression with missing cell or element TODO: Implement proper
    #   Expression handling in UFL and get rid of this
    # - Constant with missing cell TODO: Fix anything that needs to be
    #   worked around to drop this requirement

    # Extract all elements and include subelements of mixed elements
    elements = [obj.ufl_element() for obj in chain(form.arguments(), form.coefficients())]
    elements = extract_sub_elements(elements)

    # Try to find a common degree for elements
    common_degree = _auto_select_degree(elements)

    # Compute element map
    element_mapping = {}
    for element in elements:
        # Flag for whether element needs to be reconstructed
        reconstruct = False

        # Set cell
        cell = element.cell
        if cell is None:
            domains = form.ufl_domains()
            if not all(domains[0].ufl_cell() == d.ufl_cell() for d in domains):
                raise ValueError(
                    "Cannot replace unknown element cell without unique common cell in form."
                )
            cell = domains[0].ufl_cell()
            reconstruct = True

        # Set degree
        degree = element.embedded_superdegree
        if degree is None:
            degree = common_degree
            reconstruct = True

        # Reconstruct element and add to map
        if reconstruct:
            element_mapping[element] = element.reconstruct(cell=cell, degree=degree)
        else:
            element_mapping[element] = element

    return element_mapping


def _compute_max_subdomain_ids(integral_data):
    """Compute the maximum subdomain ids."""
    max_subdomain_ids = {}
    for itg_data in integral_data:
        it = itg_data.integral_type
        for integral in itg_data.integrals:
            # Convert string for default integral to -1
            sids = (-1 if isinstance(si, str) else si for si in integral.subdomain_id())
            newmax = max(sids) + 1
            prevmax = max_subdomain_ids.get(it, 0)
            max_subdomain_ids[it] = max(prevmax, newmax)
    return max_subdomain_ids


def _compute_form_data_elements(self, arguments, coefficients, domains):
    """Compute form data elements."""
    self.argument_elements = tuple(f.ufl_element() for f in arguments)
    self.coefficient_elements = tuple(f.ufl_element() for f in coefficients)
    self.coordinate_elements = tuple(domain.ufl_coordinate_element() for domain in domains)

    # TODO: Include coordinate elements from argument and coefficient
    # domains as well? Can they differ?

    # Note: Removed self.elements and self.sub_elements to make sure
    #       code that depends on the selection of argument +
    #       coefficient elements blow up, as opposed to silently
    #       almost working, with the introduction of the coordinate
    #       elements here.

    all_elements = self.argument_elements + self.coefficient_elements + self.coordinate_elements
    all_sub_elements = extract_sub_elements(all_elements)

    self.unique_elements = unique_tuple(all_elements)
    self.unique_sub_elements = unique_tuple(all_sub_elements)


def _check_elements(form_data):
    """Check elements."""
    for element in chain(form_data.unique_elements, form_data.unique_sub_elements):
        if element.cell is None:
            raise ValueError(f"Found element with undefined cell: {element}")


def _check_facet_geometry(integral_data):
    """Check facet geometry."""
    for itg_data in integral_data:
        for itg in itg_data.integrals:
            it = itg_data.integral_type
            # Facet geometry is only valid in facet integrals.
            # Allowing custom integrals to pass as well, although
            # that's not really strict enough.
            if not ("facet" in it or "custom" in it or "interface" in it):
                # Not a facet integral
                for expr in traverse_unique_terminals(itg.integrand()):
                    cls = expr._ufl_class_
                    if issubclass(cls, GeometricFacetQuantity):
                        raise ValueError(f"Integral of type {it} cannot contain a {cls.__name__}.")


def _check_form_arity(preprocessed_form):
    """Check that we don't have a mixed linear/bilinear form or anything like that."""
    # FIXME: This is slooow and should be moved to form compiler
    # and/or replaced with something faster
    if 1 != len(compute_form_arities(preprocessed_form)):
        raise ValueError("All terms in form must have same rank.")


def _build_coefficient_replace_map(coefficients, element_mapping=None):
    """Create new Coefficient objects with count starting at 0.

    Returns:
        mapping from old to new objects, and lists of the new objects
    """
    if element_mapping is None:
        element_mapping = {}

    new_coefficients = []
    replace_map = {}
    for i, f in enumerate(coefficients):
        old_e = f.ufl_element()
        new_e = element_mapping.get(old_e, old_e)
        # XXX: This is a hack to ensure that if the original
        # coefficient had a domain, the new one does too.
        # This should be overhauled with requirement that Expressions
        # always have a domain.
        domain = extract_unique_domain(f)
        if domain is not None:
            new_e = FunctionSpace(domain, new_e)
        new_f = Coefficient(new_e, count=i)
        new_coefficients.append(new_f)
        replace_map[f] = new_f

    return new_coefficients, replace_map


def attach_estimated_degrees(form):
    """Attach estimated polynomial degree to a form's integrals.

    Args:
        form: The Form` to inspect.

    Returns:
        A new Form with estimate degrees attached.
    """
    integrals = form.integrals()

    new_integrals = []
    for integral in integrals:
        md = {}
        md.update(integral.metadata())
        degree = estimate_total_polynomial_degree(integral.integrand())
        md["estimated_polynomial_degree"] = degree
        new_integrals.append(integral.reconstruct(metadata=md))
    return Form(new_integrals)


def preprocess_form(form, complex_mode):
    """Preprocess a form."""
    # Note: Default behaviour here will process form the way that is
    # currently expected by vanilla FFC

    # Check that the form does not try to compare complex quantities:
    # if the quantites being compared are 'provably' real, wrap them
    # with Real, otherwise throw an error.
    if complex_mode:
        form = do_comparison_check(form)

    # Lower abstractions for tensor-algebra types into index notation,
    # reducing the number of operators later algorithms and form
    # compilers need to handle
    form = apply_algebra_lowering(form)

    # After lowering to index notation, remove any complex nodes that
    # have been introduced but are not wanted when working in real mode,
    # allowing for purely real forms to be written
    if not complex_mode:
        form = remove_complex_nodes(form)

    # Apply differentiation before function pullbacks, because for
    # example coefficient derivatives are more complicated to derive
    # after coefficients are rewritten, and in particular for
    # user-defined coefficient relations it just gets too messy
    form = apply_derivatives(form)

    return form


def compute_form_data(
    form,
    do_apply_function_pullbacks=False,
    do_apply_integral_scaling=False,
    do_apply_geometry_lowering=False,
    preserve_geometry_types=(),
    do_apply_default_restrictions=True,
    do_apply_restrictions=True,
    do_estimate_degrees=True,
    do_append_everywhere_integrals=True,
    complex_mode=False,
):
    """Compute form data.

    The default arguments configured to behave the way old FFC expects.
    """
    # TODO: Move this to the constructor instead
    self = FormData()

    # --- Store untouched form for reference.
    # The user of FormData may get original arguments,
    # original coefficients, and form signature from this object.
    # But be aware that the set of original coefficients are not
    # the same as the ones used in the final UFC form.
    # See 'reduced_coefficients' below.
    self.original_form = form

    # --- Pass form integrands through some symbolic manipulation

    form = preprocess_form(form, complex_mode)

    # --- Group form integrals
    # TODO: Refactor this, it's rather opaque what this does
    # TODO: Is self.original_form.ufl_domains() right here?
    #       It will matter when we start including 'num_domains' in ufc form.
    form = group_form_integrals(
        form,
        self.original_form.ufl_domains(),
        do_append_everywhere_integrals=do_append_everywhere_integrals,
    )

    # Estimate polynomial degree of integrands now, before applying
    # any pullbacks and geometric lowering.  Otherwise quad degrees
    # blow up horrifically.
    if do_estimate_degrees:
        form = attach_estimated_degrees(form)

    if do_apply_function_pullbacks:
        # Rewrite coefficients and arguments in terms of their
        # reference cell values with Piola transforms and symmetry
        # transforms injected where needed.
        # Decision: Not supporting grad(dolfin.Expression) without a
        #           Domain.  Current dolfin works if Expression has a
        #           cell but this should be changed to a mesh.
        form = apply_function_pullbacks(form)

    # Scale integrals to reference cell frames
    if do_apply_integral_scaling:
        form = apply_integral_scaling(form)

    # Apply default restriction to fully continuous terminals
    if do_apply_default_restrictions:
        form = apply_default_restrictions(form)

    # Lower abstractions for geometric quantities into a smaller set
    # of quantities, allowing the form compiler to deal with a smaller
    # set of types and treating geometric quantities like any other
    # expressions w.r.t. loop-invariant code motion etc.
    if do_apply_geometry_lowering:
        form = apply_geometry_lowering(form, preserve_geometry_types)

    # Apply differentiation again, because the algorithms above can
    # generate new derivatives or rewrite expressions inside
    # derivatives
    if do_apply_function_pullbacks or do_apply_geometry_lowering:
        form = apply_derivatives(form)

        # Neverending story: apply_derivatives introduces new Jinvs,
        # which needs more geometry lowering
        if do_apply_geometry_lowering:
            form = apply_geometry_lowering(form, preserve_geometry_types)
            # Lower derivatives that may have appeared
            form = apply_derivatives(form)

    form = apply_coordinate_derivatives(form)

    # Propagate restrictions to terminals
    if do_apply_restrictions:
        form = apply_restrictions(form)

    # If in real mode, remove any complex nodes introduced during form processing.
    if not complex_mode:
        form = remove_complex_nodes(form)

    # --- Group integrals into IntegralData objects
    # Most of the heavy lifting is done above in group_form_integrals.
    self.integral_data = build_integral_data(form.integrals())

    # --- Create replacements for arguments and coefficients

    # Figure out which form coefficients each integral should enable
    for itg_data in self.integral_data:
        itg_coeffs = set()
        # Get all coefficients in integrand
        for itg in itg_data.integrals:
            itg_coeffs.update(extract_coefficients(itg.integrand()))
        # Store with IntegralData object
        itg_data.integral_coefficients = itg_coeffs

    # Figure out which coefficients from the original form are
    # actually used in any integral (Differentiation may reduce the
    # set of coefficients w.r.t. the original form)
    reduced_coefficients_set = set()
    for itg_data in self.integral_data:
        reduced_coefficients_set.update(itg_data.integral_coefficients)
    self.reduced_coefficients = sorted(reduced_coefficients_set, key=lambda c: c.count())
    self.num_coefficients = len(self.reduced_coefficients)
    self.original_coefficient_positions = [
        i for i, c in enumerate(self.original_form.coefficients()) if c in self.reduced_coefficients
    ]

    # Store back into integral data which form coefficients are used
    # by each integral
    for itg_data in self.integral_data:
        itg_data.enabled_coefficients = [
            bool(coeff in itg_data.integral_coefficients) for coeff in self.reduced_coefficients
        ]

    # --- Collect some trivial data

    # Get rank of form from argument list (assuming not a mixed arity form)
    self.rank = len(self.original_form.arguments())

    # Extract common geometric dimension (topological is not common!)
    self.geometric_dimension = self.original_form.integrals()[0].ufl_domain().geometric_dimension()

    # --- Build mapping from old incomplete element objects to new
    # well defined elements.  This is to support the Expression
    # construct in dolfin which subclasses Coefficient but doesn't
    # provide an element, and the Constant construct that doesn't
    # provide the domain that a Coefficient is supposed to have. A
    # future design iteration in UFL/UFC/FFC/DOLFIN may allow removal
    # of this mapping with the introduction of UFL types for
    # Expression-like functions that can be evaluated in quadrature
    # points.
    self.element_replace_map = _compute_element_mapping(self.original_form)

    # Mappings from elements and coefficients that reside in form to
    # objects with canonical numbering as well as completed cells and
    # elements
    renumbered_coefficients, function_replace_map = _build_coefficient_replace_map(
        self.reduced_coefficients, self.element_replace_map
    )
    self.function_replace_map = function_replace_map

    # --- Store various lists of elements and sub elements (adds
    #     members to self)
    _compute_form_data_elements(
        self,
        self.original_form.arguments(),
        renumbered_coefficients,
        self.original_form.ufl_domains(),
    )

    # --- Store number of domains for integral types
    # TODO: Group this by domain first. For now keep a backwards
    # compatible data structure.
    self.max_subdomain_ids = _compute_max_subdomain_ids(self.integral_data)

    # --- Checks
    _check_elements(self)
    _check_facet_geometry(self.integral_data)

    # TODO: This is a very expensive check... Replace with something
    # faster!
    preprocessed_form = reconstruct_form_from_integral_data(self.integral_data)

    # TODO: Test how fast this is
    check_form_arity(preprocessed_form, self.original_form.arguments(), complex_mode)

    # TODO: This member is used by unit tests, change the tests to
    # remove this!
    self.preprocessed_form = preprocessed_form

    return self



# Copyright (C) 2007 Anders Logg
#
# This file is part of UFL.
#
# UFL is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UFL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with UFL. If not, see <http://www.gnu.org/licenses/>.
#
# A mixed element of mixed elements
from ufl import triangle
from ufl.finiteelement import FiniteElement, MixedElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1

P3 = FiniteElement("Lagrange", triangle, 3, (), identity_pullback, H1)

element = MixedElement([[P3, P3], [P3, P3]])



"""This module defines the Terminal class.

Terminal the superclass for all types that are terminal nodes in an expression tree.
"""
# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Anders Logg, 2008
# Modified by Massimiliano Leoni, 2016

import warnings

from ufl.core.expr import Expr
from ufl.core.ufl_type import ufl_type


@ufl_type(is_abstract=True, is_terminal=True)
class Terminal(Expr):
    """Base class for terminal objects.

    A terminal node in the UFL expression tree.
    """

    __slots__ = ()

    def __init__(self):
        """Initialise the terminal."""
        Expr.__init__(self)

    ufl_operands = ()
    ufl_free_indices = ()
    ufl_index_dimensions = ()

    def ufl_domains(self):
        """Return tuple of domains related to this terminal object."""
        raise NotImplementedError("Missing implementation of domains().")

    def evaluate(self, x, mapping, component, index_values, derivatives=()):
        """Get *self* from *mapping* and return the component asked for."""
        f = mapping.get(self)
        # No mapping, trying to evaluate self as a constant
        if f is None:
            try:
                try:
                    f = float(self)
                except TypeError:
                    f = complex(self)
                if derivatives:
                    f = 0.0
                return f
            except Exception:
                pass
            # If it has an ufl_evaluate function, call it
            if hasattr(self, "ufl_evaluate"):
                return self.ufl_evaluate(x, component, derivatives)
            # Take component if any
            warnings.warn(
                f"Couldn't map '{self}' to a float, returning ufl object without evaluation."
            )
            f = self
            if component:
                f = f[component]
            return f

        # Found a callable in the mapping
        if callable(f):
            if derivatives:
                f = f(x, derivatives)
            else:
                f = f(x)
        else:
            if derivatives:
                return 0.0

        # Take component if any (expecting nested tuple)
        for c in component:
            f = f[c]
        return f

    def _ufl_signature_data_(self, renumbering):
        """Default signature data for of terminals just return the repr string."""
        return repr(self)

    def _ufl_compute_hash_(self):
        """Default hash of terminals just hash the repr string."""
        return hash(repr(self))

    def __eq__(self, other):
        """Default comparison of terminals just compare repr strings."""
        return repr(self) == repr(other)


# --- Subgroups of terminals ---


@ufl_type(is_abstract=True)
class FormArgument(Terminal):
    """An abstract class for a form argument (a thing in a primal finite element space)."""

    __slots__ = ()

    def __init__(self):
        """Initialise the form argument."""
        Terminal.__init__(self)



import pytest

from ufl import Index, Mesh, SpatialCoordinate, outer, triangle
from ufl.classes import FixedIndex, Indexed, MultiIndex, Outer, Zero
from ufl.finiteelement import FiniteElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1


@pytest.fixture
def domain():
    return Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))


@pytest.fixture
def x1(domain):
    x = SpatialCoordinate(domain)
    return x


@pytest.fixture
def x2(domain):
    x = SpatialCoordinate(domain)
    return outer(x, x)


@pytest.fixture
def x3(domain):
    x = SpatialCoordinate(domain)
    return outer(outer(x, x), x)


def test_annotated_literals():
    z = Zero(())
    assert z.ufl_shape == ()
    assert z.ufl_free_indices == ()
    assert z.ufl_index_dimensions == ()

    z = Zero((3,))
    assert z.ufl_shape == (3,)
    assert z.ufl_free_indices == ()
    assert z.ufl_index_dimensions == ()

    i = Index(count=2)
    j = Index(count=4)
    z = Zero((), (j, i), {i: 3, j: 5})
    assert z.ufl_shape == ()
    assert z.ufl_free_indices == (2, 4)
    assert z.ufl_index_dimensions == (3, 5)


def test_fixed_indexing_of_expression(x1, x2, x3):
    x0 = x1[0]
    x00 = x2[0, 0]
    x000 = x3[0, 0, 0]
    assert isinstance(x0, Indexed)
    assert isinstance(x00, Indexed)
    assert isinstance(x000, Indexed)
    assert isinstance(x0.ufl_operands[0], SpatialCoordinate)
    assert isinstance(x00.ufl_operands[0], Outer)
    assert isinstance(x000.ufl_operands[0], Outer)
    assert isinstance(x0.ufl_operands[1], MultiIndex)
    assert isinstance(x00.ufl_operands[1], MultiIndex)
    assert isinstance(x000.ufl_operands[1], MultiIndex)

    mi = x000.ufl_operands[1]
    assert len(mi) == 3
    assert mi.indices() == (FixedIndex(0),) * 3



"""UFL: The Unified Form Language.

The Unified Form Language is an embedded domain specific language
for definition of variational forms intended for finite element
discretization. More precisely, it defines a fixed interface for choosing
finite element spaces and defining expressions for weak forms in a
notation close to the mathematical one.

This Python module contains the language as well as algorithms to work
with it.

* To import the language, type::

    import ufl

* To import the underlying classes an UFL expression tree is built
  from, type
  ::

    import ufl.classes

* Various algorithms for working with UFL expression trees can be
  accessed by
  ::

    import ufl.algorithms

Classes and algorithms are considered implementation details and
should not be used in form definitions.

For more details on the language, see

  http://www.fenicsproject.org

and

  http://arxiv.org/abs/1211.4047

The development version can be found in the repository at

  https://github.com/FEniCS/ufl

A very brief overview of the language contents follows:

* Cells::

    -AbstractCell
    -Cell
    -TensorProductCell
    -vertex
    -interval
    -triangle
    -tetrahedron
    -quadrilateral
    -hexahedron
    -prism
    -pyramid
    -pentatope
    -tesseract

* Domains::

    -AbstractDomain
    -Mesh
    -MeshView

* Sobolev spaces::

    -L2
    -H1
    -H2
    -HInf
    -HDiv
    -HCurl
    -HEin
    -HDivDiv


* Pull backs::

    -identity_pullback
    -contravariant_piola
    -covariant_piola
    -l2_piola
    -double_contravariant_piola
    -double_covariant_piola

* Function spaces::

    -FunctionSpace
    -MixedFunctionSpace

* Arguments::

    -Argument
    -TestFunction
    -TrialFunction
    -Arguments
    -TestFunctions
    -TrialFunctions

* Coefficients::

    -Coefficient
    -Constant
    -VectorConstant
    -TensorConstant

* Splitting form arguments in mixed spaces::

    -split

* Literal constants::

    -Identity
    -PermutationSymbol

* Geometric quantities::

    -SpatialCoordinate
    -FacetNormal
    -CellNormal
    -CellVolume
    -CellDiameter
    -Circumradius
    -MinCellEdgeLength
    -MaxCellEdgeLength
    -FacetArea
    -MinFacetEdgeLength
    -MaxFacetEdgeLength
    -Jacobian
    -JacobianDeterminant
    -JacobianInverse

* Indices::

    -Index
    -indices
    -i, j, k, l
    -p, q, r, s

* Scalar to tensor expression conversion::

    -as_tensor
    -as_vector
    -as_matrix

* Unit vectors and matrices::

    -unit_vector
    -unit_vectors
    -unit_matrix
    -unit_matrices

* Tensor algebra operators::

    -outer, inner, dot, cross, perp
    -det, inv, cofac
    -transpose, tr, diag, diag_vector
    -dev, skew, sym

* Elementwise tensor operators::

    -elem_mult
    -elem_div
    -elem_pow
    -elem_op

* Differential operators::

    -variable
    (-diff,)
    -grad, nabla_grad
    -div, nabla_div
    -curl, rot
    -Dx, Dn

* Nonlinear functions::

    -max_value, min_value
    -abs, sign
    -sqrt
    -exp, ln, erf
    -cos, sin, tan
    -acos, asin, atan, atan2
    -cosh, sinh, tanh
    -bessel_J, bessel_Y, bessel_I, bessel_K

* Complex operations::

    - conj, real, imag
    conjugate is an alias for conj

* Discontinuous Galerkin operators::

    -v("+"), v("-")
    -jump
    -avg
    -cell_avg, facet_avg

* Conditional operators::

    - eq, ne, le, ge, lt, gt
    - <, >, <=, >=
    - And, Or, Not
    - conditional

* Integral measures::

    -dx, ds, dS, dP
    -dc, dC, dO, dI, dX
    -ds_b, ds_t, ds_tb, ds_v, dS_h, dS_v

* Form transformations::

    -rhs, lhs
    -system
    -functional
    -replace
    -adjoint
    -action
    (-energy_norm,)
    -sensitivity_rhs
    -derivative
"""

# Copyright (C) 2008-2016 Martin Sandve Alnæs and Anders Logg
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Kristian B. Oelgaard, 2009, 2011
# Modified by Anders Logg, 2009.
# Modified by Johannes Ring, 2014.
# Modified by Andrew T. T. McRae, 2014
# Modified by Lawrence Mitchell, 2014
# Modified by Massimiliano Leoni, 2016
# Modified by Cecile Daversin-Catty, 2018
# Modified by Nacime Bouziani, 2019

import importlib.metadata

__version__ = importlib.metadata.version("fenics-ufl")

from math import e, pi

from ufl.action import Action
from ufl.adjoint import Adjoint
from ufl.argument import (
    Argument,
    Arguments,
    Coargument,
    TestFunction,
    TestFunctions,
    TrialFunction,
    TrialFunctions,
)
from ufl.cell import AbstractCell, Cell, TensorProductCell, as_cell
from ufl.coefficient import Coefficient, Coefficients, Cofunction
from ufl.constant import Constant, TensorConstant, VectorConstant
from ufl.constantvalue import Identity, PermutationSymbol, as_ufl, zero
from ufl.core.external_operator import ExternalOperator
from ufl.core.interpolate import Interpolate, interpolate
from ufl.core.multiindex import Index, indices
from ufl.domain import AbstractDomain, Mesh, MeshView
from ufl.finiteelement import AbstractFiniteElement
from ufl.form import BaseForm, Form, FormSum, ZeroBaseForm
from ufl.formoperators import (
    action,
    adjoint,
    derivative,
    energy_norm,
    extract_blocks,
    functional,
    lhs,
    replace,
    rhs,
    sensitivity_rhs,
    system,
)
from ufl.functionspace import FunctionSpace, MixedFunctionSpace
from ufl.geometry import (
    CellDiameter,
    CellNormal,
    CellVolume,
    Circumradius,
    FacetArea,
    FacetNormal,
    Jacobian,
    JacobianDeterminant,
    JacobianInverse,
    MaxCellEdgeLength,
    MaxFacetEdgeLength,
    MinCellEdgeLength,
    MinFacetEdgeLength,
    SpatialCoordinate,
)
from ufl.integral import Integral
from ufl.matrix import Matrix
from ufl.measure import Measure, custom_integral_types, integral_types, register_integral_type
from ufl.objects import (
    dC,
    dc,
    dI,
    dO,
    dP,
    dS,
    ds,
    ds_b,
    dS_h,
    ds_t,
    ds_tb,
    dS_v,
    ds_v,
    dX,
    dx,
    facet,
    hexahedron,
    i,
    interval,
    j,
    k,
    l,
    p,
    pentatope,
    prism,
    pyramid,
    q,
    quadrilateral,
    r,
    s,
    tesseract,
    tetrahedron,
    triangle,
    vertex,
)
from ufl.operators import (
    And,
    Dn,
    Dx,
    Not,
    Or,
    acos,
    asin,
    atan,
    atan2,
    avg,
    bessel_I,
    bessel_J,
    bessel_K,
    bessel_Y,
    cell_avg,
    cofac,
    conditional,
    conj,
    cos,
    cosh,
    cross,
    curl,
    det,
    dev,
    diag,
    diag_vector,
    diff,
    div,
    dot,
    elem_div,
    elem_mult,
    elem_op,
    elem_pow,
    eq,
    erf,
    exp,
    exterior_derivative,
    facet_avg,
    ge,
    grad,
    gt,
    imag,
    inner,
    inv,
    jump,
    le,
    ln,
    lt,
    max_value,
    min_value,
    nabla_div,
    nabla_grad,
    ne,
    outer,
    perp,
    rank,
    real,
    rot,
    shape,
    sign,
    sin,
    sinh,
    skew,
    sqrt,
    sym,
    tan,
    tanh,
    tr,
    transpose,
    variable,
)
from ufl.pullback import (
    AbstractPullback,
    MixedPullback,
    SymmetricPullback,
    contravariant_piola,
    covariant_piola,
    double_contravariant_piola,
    double_covariant_piola,
    identity_pullback,
    l2_piola,
)
from ufl.sobolevspace import H1, H2, L2, HCurl, HDiv, HDivDiv, HEin, HInf
from ufl.split_functions import split
from ufl.tensors import (
    as_matrix,
    as_tensor,
    as_vector,
    unit_matrices,
    unit_matrix,
    unit_vector,
    unit_vectors,
)
from ufl.utils.sequences import product

__all__ = [
    "product",
    "as_cell",
    "AbstractCell",
    "Cell",
    "TensorProductCell",
    "AbstractDomain",
    "Mesh",
    "MeshView",
    "L2",
    "H1",
    "H2",
    "HCurl",
    "HDiv",
    "HInf",
    "HEin",
    "HDivDiv",
    "identity_pullback",
    "l2_piola",
    "contravariant_piola",
    "covariant_piola",
    "double_contravariant_piola",
    "double_covariant_piola",
    "l2_piola",
    "MixedPullback",
    "SymmetricPullback",
    "AbstractPullback",
    "SpatialCoordinate",
    "CellVolume",
    "CellDiameter",
    "Circumradius",
    "MinCellEdgeLength",
    "MaxCellEdgeLength",
    "FacetArea",
    "MinFacetEdgeLength",
    "MaxFacetEdgeLength",
    "FacetNormal",
    "CellNormal",
    "Jacobian",
    "JacobianDeterminant",
    "JacobianInverse",
    "AbstractFiniteElement",
    "FunctionSpace",
    "MixedFunctionSpace",
    "Argument",
    "Coargument",
    "TestFunction",
    "TrialFunction",
    "Arguments",
    "TestFunctions",
    "TrialFunctions",
    "Coefficient",
    "Cofunction",
    "Coefficients",
    "Matrix",
    "Adjoint",
    "Action",
    "Interpolate",
    "interpolate",
    "ExternalOperator",
    "Constant",
    "VectorConstant",
    "TensorConstant",
    "split",
    "PermutationSymbol",
    "Identity",
    "zero",
    "as_ufl",
    "Index",
    "indices",
    "as_tensor",
    "as_vector",
    "as_matrix",
    "unit_vector",
    "unit_vectors",
    "unit_matrix",
    "unit_matrices",
    "rank",
    "shape",
    "conj",
    "real",
    "imag",
    "outer",
    "inner",
    "dot",
    "cross",
    "perp",
    "det",
    "inv",
    "cofac",
    "transpose",
    "tr",
    "diag",
    "diag_vector",
    "dev",
    "skew",
    "sym",
    "sqrt",
    "exp",
    "ln",
    "erf",
    "cos",
    "sin",
    "tan",
    "acos",
    "asin",
    "atan",
    "atan2",
    "cosh",
    "sinh",
    "tanh",
    "bessel_J",
    "bessel_Y",
    "bessel_I",
    "bessel_K",
    "eq",
    "ne",
    "le",
    "ge",
    "lt",
    "gt",
    "And",
    "Or",
    "Not",
    "conditional",
    "sign",
    "max_value",
    "min_value",
    "variable",
    "diff",
    "Dx",
    "grad",
    "div",
    "curl",
    "rot",
    "nabla_grad",
    "nabla_div",
    "Dn",
    "exterior_derivative",
    "jump",
    "avg",
    "cell_avg",
    "facet_avg",
    "elem_mult",
    "elem_div",
    "elem_pow",
    "elem_op",
    "Form",
    "BaseForm",
    "FormSum",
    "ZeroBaseForm",
    "Integral",
    "Measure",
    "register_integral_type",
    "integral_types",
    "custom_integral_types",
    "replace",
    "derivative",
    "action",
    "energy_norm",
    "rhs",
    "lhs",
    "extract_blocks",
    "system",
    "functional",
    "adjoint",
    "sensitivity_rhs",
    "dx",
    "ds",
    "dS",
    "dP",
    "dc",
    "dC",
    "dO",
    "dI",
    "dX",
    "ds_b",
    "ds_t",
    "ds_tb",
    "ds_v",
    "dS_h",
    "dS_v",
    "vertex",
    "interval",
    "triangle",
    "tetrahedron",
    "prism",
    "pyramid",
    "pentatope",
    "tesseract",
    "quadrilateral",
    "hexahedron",
    "facet",
    "i",
    "j",
    "k",
    "l",
    "p",
    "q",
    "r",
    "s",
    "e",
    "pi",
]



"""UFL to unicode."""

import numbers

import ufl
from ufl.algorithms import compute_form_data
from ufl.core.multiindex import FixedIndex, Index
from ufl.corealg.map_dag import map_expr_dag
from ufl.corealg.multifunction import MultiFunction
from ufl.form import Form

try:
    import colorama

    has_colorama = True
except ImportError:
    has_colorama = False


class PrecedenceRules(MultiFunction):
    """An enum-like class for C operator precedence levels."""

    def __init__(self):
        """Initialise."""
        MultiFunction.__init__(self)

    def highest(self, o):
        """Return the highest precendence."""
        return 0

    terminal = highest
    list_tensor = highest
    component_tensor = highest

    def restricted(self, o):
        """Return precedence of a restriced."""
        return 5

    cell_avg = restricted
    facet_avg = restricted

    def call(self, o):
        """Return precedence of a call."""
        return 10

    indexed = call
    min_value = call
    max_value = call
    math_function = call
    bessel_function = call

    def power(self, o):
        """Return precedence of a power."""
        return 12

    def mathop(self, o):
        """Return precedence of a mathop."""
        return 15

    derivative = mathop
    trace = mathop
    deviatoric = mathop
    cofactor = mathop
    skew = mathop
    sym = mathop

    def not_condition(self, o):
        """Return precedence of a not_condition."""
        return 20

    def product(self, o):
        """Return precedence of a product."""
        return 30

    division = product
    # mod = product
    dot = product
    inner = product
    outer = product
    cross = product

    def add(self, o):
        """Return precedence of an add."""
        return 40

    # sub = add
    index_sum = add

    def lt(self, o):
        """Return precedence of a lt."""
        return 50

    le = lt
    gt = lt
    ge = lt

    def eq(self, o):
        """Return precedence of an eq."""
        return 60

    ne = eq

    def and_condition(self, o):
        """Return precedence of an and_condition."""
        return 70

    def or_condition(self, o):
        """Return precedence of an or_condition."""
        return 71

    def conditional(self, o):
        """Return precedence of a conditional."""
        return 72

    def lowest(self, o):
        """Return precedence of a lowest."""
        return 80

    operator = lowest


_precrules = PrecedenceRules()


def precedence(expr):
    """Get the precedence of an expr."""
    return _precrules(expr)


class UC:
    """An enum-like class for unicode characters."""

    # Letters in this alphabet have contiguous code point numbers
    bold_math_a = "𝐚"
    bold_math_A = "𝐀"

    thin_space = "\u2009"

    superscript_plus = "⁺"
    superscript_minus = "⁻"
    superscript_equals = "⁼"
    superscript_left_paren = "⁽"
    superscript_right_paren = "⁾"
    superscript_digits = ["⁰", "¹", "²", "³", "⁴", "⁵", "⁶", "⁷", "⁸", "⁹"]

    subscript_plus = "₊"
    subscript_minus = "₋"
    subscript_equals = "₌"
    subscript_left_paren = "₍"
    subscript_right_paren = "₎"
    subscript_digits = ["₀", "₁", "₂", "₃", "₄", "₅", "₆", "₇", "₈", "₉"]

    sqrt = "√"
    transpose = "ᵀ"

    integral = "∫"
    integral_double = "∬"
    integral_triple = "∭"
    integral_contour = "∮"
    integral_surface = "∯"
    integral_volume = "∰"

    sum = "∑"
    division_slash = "∕"
    partial = "∂"
    epsilon = "ε"
    omega = "ω"
    Omega = "Ω"
    gamma = "γ"
    Gamma = "Γ"
    nabla = "∇"
    for_all = "∀"

    dot = "⋅"
    cross_product = "⨯"
    circled_times = "⊗"
    nary_product = "∏"

    ne = "≠"
    lt = "<"
    le = "≤"
    gt = ">"
    ge = "≥"

    logical_and = "∧"
    logical_or = "∨"
    logical_not = "¬"

    element_of = "∈"
    not_element_of = "∉"

    left_white_square_bracket = "⟦"
    right_white_squared_bracket = "⟧"
    left_angled_bracket = "⟨"
    right_angled_bracket = "⟩"
    left_double_angled_bracket = "⟪"
    right_double_angled_bracket = "⟫"

    combining_right_arrow_above = "\u20d7"
    combining_overline = "\u0305"


def bolden_letter(c):
    """Bolden a letter."""
    if ord("A") <= ord(c) <= ord("Z"):
        c = chr(ord(c) - ord("A") + ord(UC.bold_math_A))
    elif ord("a") <= ord(c) <= ord("z"):
        c = chr(ord(c) - ord("a") + ord(UC.bold_math_a))
    return c


def superscript_digit(digit):
    """Make a digit superscript."""
    return UC.superscript_digits[ord(digit) - ord("0")]


def subscript_digit(digit):
    """Make a digit subscript."""
    return UC.subscript_digits[ord(digit) - ord("0")]


def bolden_string(s):
    """Bolden a string."""
    return "".join(bolden_letter(c) for c in s)


def overline_string(f):
    """Overline a string."""
    return "".join(f"{c}{UC.combining_overline}" for c in f)


def subscript_number(number):
    """Make a number subscript."""
    assert isinstance(number, int)
    prefix = UC.subscript_minus if number < 0 else ""
    number = str(number)
    return prefix + "".join(subscript_digit(c) for c in str(number))


def superscript_number(number):
    """Make a number superscript."""
    assert isinstance(number, int)
    prefix = UC.superscript_minus if number < 0 else ""
    number = str(number)
    return prefix + "".join(superscript_digit(c) for c in str(number))


def opfont(opname):
    """Use the font for operators."""
    return bolden_string(opname)


def measure_font(dx):
    """Use the font for measures."""
    return bolden_string(dx)


integral_by_dim = {3: UC.integral_triple, 2: UC.integral_double, 1: UC.integral, 0: UC.integral}

integral_type_to_codim = {
    "cell": 0,
    "exterior_facet": 1,
    "interior_facet": 1,
    "vertex": "tdim",
    "point": "tdim",
    "custom": 0,
    "overlap": 0,
    "interface": 1,
    "cutcell": 0,
}

integral_symbols = {
    "cell": UC.integral_volume,
    "exterior_facet": UC.integral_surface,
    "interior_facet": UC.integral_surface,
    "vertex": UC.integral,
    "point": UC.integral,
    "custom": UC.integral,
    "overlap": UC.integral,
    "interface": UC.integral,
    "cutcell": UC.integral,
}

integral_postfixes = {
    "cell": "",
    "exterior_facet": "ext",
    "interior_facet": "int",
    "vertex": "vertex",
    "point": "point",
    "custom": "custom",
    "overlap": "overlap",
    "interface": "interface",
    "cutcell": "cutcell",
}


def get_integral_symbol(integral_type, domain, subdomain_id):
    """Get the symbol for an integral."""
    tdim = domain.topological_dimension()
    codim = integral_type_to_codim[integral_type]
    itgdim = tdim - codim

    # ipost = integral_postfixes[integral_type]
    istr = integral_by_dim[itgdim]

    # TODO: Render domain description

    subdomain_strs = []
    for subdomain in subdomain_id:
        if isinstance(subdomain, numbers.Integral):
            subdomain_strs.append(subscript_number(int(subdomain)))
        elif subdomain == "everywhere":
            pass
        elif subdomain == "otherwise":
            subdomain_strs.append("[rest of domain]")
    istr += ",".join(subdomain_strs)

    dxstr = ufl.measure.integral_type_to_measure_name[integral_type]
    dxstr = measure_font(dxstr)

    return istr, dxstr


def par(s):
    """Wrap in parentheses."""
    return f"({s})"


def is_int(s):
    """Check if a value is an integer."""
    try:
        int(s)
        return True
    except ValueError:
        return False


def format_index(ii):
    """Format an index."""
    if isinstance(ii, FixedIndex):
        s = f"{ii._value}"
    elif isinstance(ii, Index):
        s = "i{subscript_number(ii._count)}"
    else:
        raise ValueError(f"Invalid index type {type(ii)}.")
    return s


def ufl2unicode(expression):
    """Generate Unicode string for a UFL expression or form."""
    if isinstance(expression, Form):
        form_data = compute_form_data(expression)
        preprocessed_form = form_data.preprocessed_form
        return form2unicode(preprocessed_form, form_data)
    else:
        return expression2unicode(ufl.as_ufl(expression))


def expression2unicode(expression, argument_names=None, coefficient_names=None):
    """Generate Unicode string for a UFL expression."""
    rules = Expression2UnicodeHandler(argument_names, coefficient_names)
    return map_expr_dag(rules, expression)


def form2unicode(form, formdata):
    """Generate Unicode string for a UFL form."""
    argument_names = None
    coefficient_names = None

    # Define form as sum of integrals
    lines = []
    integrals = form.integrals()
    for itg in integrals:
        integrand_string = expression2unicode(itg.integrand(), argument_names, coefficient_names)

        istr, dxstr = get_integral_symbol(itg.integral_type(), itg.ufl_domain(), itg.subdomain_id())

        line = f"{istr} {integrand_string} {dxstr}"
        lines.append(line)

    return "\n  + ".join(lines)


def binop(expr, a, b, op, sep=" "):
    """Format a binary operation."""
    eprec = precedence(expr)
    op0, op1 = expr.ufl_operands
    aprec = precedence(op0)
    bprec = precedence(op1)
    # Assuming left-to-right evaluation, therefore >= and > here:
    if aprec >= eprec:
        a = par(a)
    if bprec > eprec:
        b = par(b)
    return sep.join((a, op, b))


def mathop(expr, arg, opname):
    """Format a math operation."""
    eprec = precedence(expr)
    aprec = precedence(expr.ufl_operands[0])
    op = opfont(opname)
    if aprec > eprec:
        arg = par(arg)
        sep = ""
    else:
        sep = UC.thin_space
    return f"{op}{sep}{arg}"


class Expression2UnicodeHandler(MultiFunction):
    """Convert expressions to unicode."""

    def __init__(self, argument_names=None, coefficient_names=None, colorama_bold=False):
        """Initialise."""
        MultiFunction.__init__(self)
        self.argument_names = argument_names
        self.coefficient_names = coefficient_names
        self.colorama_bold = colorama_bold and has_colorama

    # --- Terminal objects ---

    def scalar_value(self, o):
        """Format a scalar_value."""
        if o.ufl_shape and self.colorama_bold:
            return f"{colorama.Style.BRIGHT}{o._value}{colorama.Style.RESET_ALL}"
        return f"{o._value}"

    def zero(self, o):
        """Format a zero."""
        if o.ufl_shape and self.colorama_bold:
            if len(o.ufl_shape) == 1:
                return f"0{UC.combining_right_arrow_above}"
            return f"{colorama.Style.BRIGHT}0{colorama.Style.RESET_ALL}"
        return "0"

    def identity(self, o):
        """Format a identity."""
        if self.colorama_bold:
            return f"{colorama.Style.BRIGHT}I{colorama.Style.RESET_ALL}"
        return "I"

    def permutation_symbol(self, o):
        """Format a permutation_symbol."""
        if self.colorama_bold:
            return f"{colorama.Style.BRIGHT}{UC.epsilon}{colorama.Style.RESET_ALL}"
        return UC.epsilon

    def facet_normal(self, o):
        """Format a facet_normal."""
        return f"n{UC.combining_right_arrow_above}"

    def spatial_coordinate(self, o):
        """Format a spatial_coordinate."""
        return f"x{UC.combining_right_arrow_above}"

    def argument(self, o):
        """Format an argument."""
        # Using ^ for argument numbering and _ for indexing since
        # indexing is more common than exponentiation
        if self.argument_names is None:
            i = o.number()
            bfn = "v" if i == 0 else "u"
            if not o.ufl_shape:
                return bfn
            elif len(o.ufl_shape) == 1:
                return f"{bfn}{UC.combining_right_arrow_above}"
            elif self.colorama_bold:
                return f"{colorama.Style.BRIGHT}{bfn}{colorama.Style.RESET_ALL}"
            else:
                return bfn
        return self.argument_names[(o.number(), o.part())]

    def coefficient(self, o):
        """Format a coefficient."""
        # Using ^ for coefficient numbering and _ for indexing since
        # indexing is more common than exponentiation
        if self.coefficient_names is None:
            i = o.count()
            var = "w"
            if len(o.ufl_shape) == 1:
                var += UC.combining_right_arrow_above
            elif len(o.ufl_shape) > 1 and self.colorama_bold:
                var = f"{colorama.Style.BRIGHT}{var}{colorama.Style.RESET_ALL}"
            return f"{var}{subscript_number(i)}"
        return self.coefficient_names[o.count()]

    def cofunction(self, o):
        """Format a cofunction."""
        if self.coefficient_names is None:
            i = o.count()
            var = "cofunction"
            if len(o.ufl_shape) == 1:
                var += UC.combining_right_arrow_above
            elif len(o.ufl_shape) > 1 and self.colorama_bold:
                var = f"{colorama.Style.BRIGHT}{var}{colorama.Style.RESET_ALL}"
            return f"{var}{subscript_number(i)}"
        return self.coefficient_names[o.count()]

    def base_form_operator(self, o):
        """Format a base_form_operator."""
        return "BaseFormOperator"

    def action(self, o, a, b):
        """Format an Action."""
        return f"Action({a}, {b})"

    def constant(self, o):
        """Format a constant."""
        i = o.count()
        var = "c"
        if len(o.ufl_shape) == 1:
            var += UC.combining_right_arrow_above
        elif len(o.ufl_shape) > 1 and self.colorama_bold:
            var = f"{colorama.Style.BRIGHT}{var}{colorama.Style.RESET_ALL}"
        return f"{var}{superscript_number(i)}"

    def multi_index(self, o):
        """Format a multi_index."""
        return ",".join(format_index(i) for i in o)

    def label(self, o):
        """Format a label."""
        return f"l{subscript_number(o.count())}"

    # --- Non-terminal objects ---

    def variable(self, o, f, a):
        """Format a variable."""
        return f"var({f},{a})"

    def index_sum(self, o, f, i):
        """Format a index_sum."""
        if 1:  # prec(o.ufl_operands[0]) >? prec(o):
            f = par(f)
        return f"{UC.sum}[{i}]{f}"

    def sum(self, o, a, b):
        """Format a sum."""
        return binop(o, a, b, "+")

    def product(self, o, a, b):
        """Format a product."""
        return binop(o, a, b, " ", sep="")

    def division(self, o, a, b):
        """Format a division."""
        if is_int(b):
            b = subscript_number(int(b))
            if is_int(a):
                # Return as a fraction
                # NOTE: Maybe consider using fractional slash
                #  with normal numbers if terminals can handle it
                a = superscript_number(int(a))
            else:
                a = par(a)
            return f"{a} {UC.division_slash} {b}"
        return binop(o, a, b, UC.division_slash)

    def abs(self, o, a):
        """Format an ans."""
        return f"|{a}|"

    def transposed(self, o, a):
        """Format a transposed."""
        a = par(a)
        return f"{a}{UC.transpose}"

    def indexed(self, o, A, ii):
        """Format an indexed."""
        op0, op1 = o.ufl_operands
        Aprec = precedence(op0)
        oprec = precedence(o)
        if Aprec > oprec:
            A = par(A)
        return f"{A}[{ii}]"

    def variable_derivative(self, o, f, v):
        """Format a variable_derivative."""
        f = par(f)
        v = par(v)
        nom = f"{UC.partial}{f}"
        denom = f"{UC.partial}{v}"
        return par(f"{nom}{UC.division_slash}{denom}")

    def coefficient_derivative(self, o, f, w, v, cd):
        """Format a coefficient_derivative."""
        f = par(f)
        w = par(w)
        nom = f"{UC.partial}{f}"
        denom = f"{UC.partial}{w}"
        return par(f"{nom}{UC.division_slash}{denom}[{v}]")

    def grad(self, o, f):
        """Format a grad."""
        return mathop(o, f, "grad")

    def div(self, o, f):
        """Format a div."""
        return mathop(o, f, "div")

    def nabla_grad(self, o, f):
        """Format a nabla_grad."""
        oprec = precedence(o)
        fprec = precedence(o.ufl_operands[0])
        if fprec > oprec:
            f = par(f)
        return f"{UC.nabla}{UC.thin_space}{f}"

    def nabla_div(self, o, f):
        """Format a nabla_div."""
        oprec = precedence(o)
        fprec = precedence(o.ufl_operands[0])
        if fprec > oprec:
            f = par(f)
        return f"{UC.nabla}{UC.thin_space}{UC.dot}{UC.thin_space}{f}"

    def curl(self, o, f):
        """Format a curl."""
        oprec = precedence(o)
        fprec = precedence(o.ufl_operands[0])
        if fprec > oprec:
            f = par(f)
        return f"{UC.nabla}{UC.thin_space}{UC.cross_product}{UC.thin_space}{f}"

    def math_function(self, o, f):
        """Format a math_function."""
        op = opfont(o._name)
        return f"{op}{par(f)}"

    def sqrt(self, o, f):
        """Format a sqrt."""
        return f"{UC.sqrt}{par(f)}"

    def exp(self, o, f):
        """Format a exp."""
        op = opfont("exp")
        return f"{op}{par(f)}"

    def atan2(self, o, f1, f2):
        """Format a atan2."""
        f1 = par(f1)
        f2 = par(f2)
        op = opfont("arctan2")
        return f"{op}({f1}, {f2})"

    def bessel_j(self, o, nu, f):
        """Format a bessel_j."""
        op = opfont("J")
        nu = subscript_number(int(nu))
        return f"{op}{nu}{par(f)}"

    def bessel_y(self, o, nu, f):
        """Format a bessel_y."""
        op = opfont("Y")
        nu = subscript_number(int(nu))
        return f"{op}{nu}{par(f)}"

    def bessel_i(self, o, nu, f):
        """Format a bessel_i."""
        op = opfont("I")
        nu = subscript_number(int(nu))
        return f"{op}{nu}{par(f)}"

    def bessel_K(self, o, nu, f):
        """Format a bessel_K."""
        op = opfont("K")
        nu = subscript_number(int(nu))
        return f"{op}{nu}{par(f)}"

    def power(self, o, a, b):
        """Format a power."""
        if is_int(b):
            b = superscript_number(int(b))
            return binop(o, a, b, "", sep="")
        return binop(o, a, b, "^", sep="")

    def outer(self, o, a, b):
        """Format an outer."""
        return binop(o, a, b, UC.circled_times)

    def inner(self, o, a, b):
        """Format an inner."""
        return f"{UC.left_angled_bracket}{a}, {b}{UC.right_angled_bracket}"

    def dot(self, o, a, b):
        """Format a dot."""
        return binop(o, a, b, UC.dot)

    def cross(self, o, a, b):
        """Format a cross."""
        return binop(o, a, b, UC.cross_product)

    def determinant(self, o, A):
        """Format a determinant."""
        return f"|{A}|"

    def inverse(self, o, A):
        """Format an inverse."""
        A = par(A)
        return f"{A}{superscript_number(-1)}"

    def trace(self, o, A):
        """Format a trace."""
        return mathop(o, A, "tr")

    def deviatoric(self, o, A):
        """Format a deviatoric."""
        return mathop(o, A, "dev")

    def cofactor(self, o, A):
        """Format a cofactor."""
        return mathop(o, A, "cofac")

    def skew(self, o, A):
        """Format a skew."""
        return mathop(o, A, "skew")

    def sym(self, o, A):
        """Format a sym."""
        return mathop(o, A, "sym")

    def conj(self, o, a):
        """Format a conj."""
        # Overbar is already taken for average, and there is no superscript asterix in unicode.
        return mathop(o, a, "conj")

    def real(self, o, a):
        """Format a real."""
        return mathop(o, a, "Re")

    def imag(self, o, a):
        """Format a imag."""
        return mathop(o, a, "Im")

    def list_tensor(self, o, *ops):
        """Format a list_tensor."""
        return f"[{', '.join(ops)}]"

    def component_tensor(self, o, A, ii):
        """Format a component_tensor."""
        return f"[{A} {UC.for_all} {ii}]"

    def positive_restricted(self, o, f):
        """Format a positive_restriced."""
        return f"{par(f)}{UC.superscript_plus}"

    def negative_restricted(self, o, f):
        """Format a negative_restriced."""
        return f"{par(f)}{UC.superscript_minus}"

    def cell_avg(self, o, f):
        """Format a cell_avg."""
        f = overline_string(f)
        return f

    def facet_avg(self, o, f):
        """Format a facet_avg."""
        f = overline_string(f)
        return f

    def eq(self, o, a, b):
        """Format an eq."""
        return binop(o, a, b, "=")

    def ne(self, o, a, b):
        """Format a ne."""
        return binop(o, a, b, UC.ne)

    def le(self, o, a, b):
        """Format a le."""
        return binop(o, a, b, UC.le)

    def ge(self, o, a, b):
        """Format a ge."""
        return binop(o, a, b, UC.ge)

    def lt(self, o, a, b):
        """Format a lt."""
        return binop(o, a, b, UC.lt)

    def gt(self, o, a, b):
        """Format a gt."""
        return binop(o, a, b, UC.gt)

    def and_condition(self, o, a, b):
        """Format an and_condition."""
        return binop(o, a, b, UC.logical_and)

    def or_condition(self, o, a, b):
        """Format an or_condition."""
        return binop(o, a, b, UC.logical_or)

    def not_condition(self, o, a):
        """Format a not_condition."""
        a = par(a)
        return f"{UC.logical_not}{a}"

    def conditional(self, o, c, t, f):
        """Format a conditional."""
        c = par(c)
        t = par(t)
        f = par(t)
        If = opfont("if")
        Else = opfont("else")
        return f"{t} {If} {c} {Else} {f}"

    def min_value(self, o, a, b):
        """Format an min_value."""
        op = opfont("min")
        return f"{op}({a}, {b})"

    def max_value(self, o, a, b):
        """Format an max_value."""
        op = opfont("max")
        return f"{op}({a}, {b})"

    def expr_list(self, o, *ops):
        """Format an expr_list."""
        items = ", ".join(ops)
        return f"{UC.left_white_square_bracket} {items} {UC.right_white_squared_bracket}"

    def expr_mapping(self, o, *ops):
        """Format an expr_mapping."""
        items = ", ".join(ops)
        return f"{UC.left_double_angled_bracket} {items} {UC.left_double_angled_bracket}"

    def expr(self, o):
        """Format an expr."""
        raise ValueError(f"Missing handler for type {type(o)}")



__authors__ = "Martin Sandve Alnæs"
__date__ = "2009-03-14 -- 2009-03-14"

from ufl import Coefficient, FunctionSpace, Mesh, TestFunction, as_vector, product, split, triangle
from ufl.finiteelement import FiniteElement, MixedElement, SymmetricElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1


def test_split(self):
    cell = triangle
    d = cell.topological_dimension()
    domain = Mesh(FiniteElement("Lagrange", cell, 1, (d,), identity_pullback, H1))
    f = FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1)
    v = FiniteElement(
        "Lagrange", cell, 1, (d,), identity_pullback, H1, sub_elements=[f for _ in range(d)]
    )
    w = FiniteElement(
        "Lagrange", cell, 1, (d + 1,), identity_pullback, H1, sub_elements=[f for _ in range(d + 1)]
    )
    t = FiniteElement(
        "Lagrange", cell, 1, (d, d), identity_pullback, H1, sub_elements=[f for _ in range(d**2)]
    )
    s = SymmetricElement({(0, 0): 0, (0, 1): 1, (1, 0): 1, (1, 1): 2}, [f for _ in range(3)])
    m = MixedElement([f, v, w, t, s, s])

    f_space = FunctionSpace(domain, f)
    v_space = FunctionSpace(domain, v)
    w_space = FunctionSpace(domain, w)
    t_space = FunctionSpace(domain, t)
    s_space = FunctionSpace(domain, s)
    m_space = FunctionSpace(domain, m)

    # Check that shapes of all these functions are correct:
    assert () == Coefficient(f_space).ufl_shape
    assert (d,) == Coefficient(v_space).ufl_shape
    assert (d + 1,) == Coefficient(w_space).ufl_shape
    assert (d, d) == Coefficient(t_space).ufl_shape
    assert (d, d) == Coefficient(s_space).ufl_shape
    # sum of value sizes, not accounting for symmetries:
    assert (3 * d * d + 2 * d + 2,) == Coefficient(m_space).ufl_shape

    # Shapes of subelements are reproduced:
    g = Coefficient(m_space)
    (s,) = g.ufl_shape
    for g2 in split(g):
        s -= product(g2.ufl_shape)
    assert s == 0

    # Mixed elements of non-scalar subelements are flattened
    v2 = MixedElement([v, v])
    m2 = MixedElement([t, t])
    v2_space = FunctionSpace(domain, v2)
    m2_space = FunctionSpace(domain, m2)
    # assert d == 2
    # assert (2,2) == Coefficient(v2_space).ufl_shape
    assert (d + d,) == Coefficient(v2_space).ufl_shape
    assert (2 * d * d,) == Coefficient(m2_space).ufl_shape

    # Split simple element
    t = TestFunction(f_space)
    assert split(t) == (t,)

    # Split twice on nested mixed elements gets
    # the innermost scalar subcomponents
    t = TestFunction(FunctionSpace(domain, MixedElement([f, v])))
    assert split(t) == (t[0], as_vector((t[1], t[2])))
    assert split(split(t)[1]) == (t[1], t[2])
    t = TestFunction(FunctionSpace(domain, MixedElement([f, [f, v]])))
    assert split(t) == (t[0], as_vector((t[1], t[2], t[3])))
    assert split(split(t)[1]) == (t[1], as_vector((t[2], t[3])))
    t = TestFunction(FunctionSpace(domain, MixedElement([[v, f], [f, v]])))
    assert split(t) == (as_vector((t[0], t[1], t[2])), as_vector((t[3], t[4], t[5])))
    assert split(split(t)[0]) == (as_vector((t[0], t[1])), t[2])
    assert split(split(t)[1]) == (t[3], as_vector((t[4], t[5])))
    assert split(split(split(t)[0])[0]) == (t[0], t[1])
    assert split(split(split(t)[0])[1]) == (t[2],)
    assert split(split(split(t)[1])[0]) == (t[3],)
    assert split(split(split(t)[1])[1]) == (t[4], t[5])



"""Balancing."""
# -*- coding: utf-8 -*-
# Copyright (C) 2011-2017 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

from ufl.classes import (
    CellAvg,
    FacetAvg,
    Grad,
    Indexed,
    NegativeRestricted,
    PositiveRestricted,
    ReferenceGrad,
    ReferenceValue,
)
from ufl.corealg.map_dag import map_expr_dag
from ufl.corealg.multifunction import MultiFunction

modifier_precedence = [
    ReferenceValue,
    ReferenceGrad,
    Grad,
    CellAvg,
    FacetAvg,
    PositiveRestricted,
    NegativeRestricted,
    Indexed,
]

modifier_precedence = {m._ufl_handler_name_: i for i, m in enumerate(modifier_precedence)}


def balance_modified_terminal(expr):
    """Balance modified terminal."""
    # NB! Assuming e.g. grad(cell_avg(expr)) does not occur,
    # i.e. it is simplified to 0 immediately.

    if expr._ufl_is_terminal_:
        return expr

    assert expr._ufl_is_terminal_modifier_

    orig = expr

    # Build list of modifier layers
    layers = [expr]
    while not expr._ufl_is_terminal_:
        assert expr._ufl_is_terminal_modifier_
        expr = expr.ufl_operands[0]
        layers.append(expr)
    assert layers[-1] is expr
    assert expr._ufl_is_terminal_

    # Apply modifiers in order
    layers = sorted(layers[:-1], key=lambda e: modifier_precedence[e._ufl_handler_name_])
    for op in layers:
        ops = (expr,) + op.ufl_operands[1:]
        expr = op._ufl_expr_reconstruct_(*ops)

    # Preserve id if nothing has changed
    return orig if expr == orig else expr


class BalanceModifiers(MultiFunction):
    """Balance modifiers."""

    def expr(self, expr, *ops):
        """Apply to expr."""
        return expr._ufl_expr_reconstruct_(*ops)

    def terminal(self, expr):
        """Apply to terminal."""
        return expr

    def _modifier(self, expr, *ops):
        """Apply to _modifier."""
        return balance_modified_terminal(expr)

    reference_value = _modifier
    reference_grad = _modifier
    grad = _modifier
    cell_avg = _modifier
    facet_avg = _modifier
    positive_restricted = _modifier
    negative_restricted = _modifier


def balance_modifiers(expr):
    """Balance modifiers."""
    mf = BalanceModifiers()
    return map_expr_dag(mf, expr)



#
# Author: Martin Sandve Alnes
# Date: 2008-10-03
#
from ufl import Coefficient, Constant, FunctionSpace, Mesh, derivative, dx, exp, interval, variable
from ufl.finiteelement import FiniteElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1

cell = interval
element = FiniteElement("Lagrange", cell, 2, (), identity_pullback, H1)
domain = Mesh(FiniteElement("Lagrange", cell, 1, (1,), identity_pullback, H1))
space = FunctionSpace(domain, element)

u = Coefficient(space)
b = Constant(domain)
K = Constant(domain)

E = u.dx(0) + u.dx(0) ** 2 / 2
E = variable(E)
Q = b * E**2
psi = K * (exp(Q) - 1)

f = psi * dx
F = derivative(f, u)
J = derivative(-F, u)

forms = [f, F, J]



"""This module provides basic mathematical functions."""
# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Anders Logg, 2008
# Modified by Kristian B. Oelgaard, 2011

import cmath
import math
import numbers
import warnings

from ufl.constantvalue import (
    ComplexValue,
    ConstantValue,
    FloatValue,
    IntValue,
    RealValue,
    Zero,
    as_ufl,
    is_true_ufl_scalar,
)
from ufl.core.operator import Operator
from ufl.core.ufl_type import ufl_type

"""
TODO: Include additional functions available in <cmath> (need derivatives as well):

Exponential and logarithmic functions:
log10    Compute common logarithm (function)

TODO: Any other useful special functions?

About bessel functions:
http://en.wikipedia.org/wiki/Bessel_function

Portable implementations of bessel functions:
http://www.boost.org/doc/libs/1_47_0/libs/math/doc/sf_and_dist/html/math_toolkit/main_overview/tr1.html

Implementation in C++ std::tr1:: or boost::math::tr1::
- BesselK: cyl_bessel_k(nu, x)
- BesselI: cyl_bessel_i(nu, x)
- BesselJ: cyl_bessel_j(nu, x)
- BesselY: cyl_neumann(nu, x)
"""


# --- Function representations ---


@ufl_type(is_abstract=True, is_scalar=True, num_ops=1)
class MathFunction(Operator):
    """Base class for all unary scalar math functions."""

    # Freeze member variables for objects in this class
    __slots__ = ("_name",)

    def __init__(self, name, argument):
        """Initialise."""
        Operator.__init__(self, (argument,))
        if not is_true_ufl_scalar(argument):
            raise ValueError("Expecting scalar argument.")
        self._name = name

    def evaluate(self, x, mapping, component, index_values):
        """Evaluate."""
        a = self.ufl_operands[0].evaluate(x, mapping, component, index_values)
        try:
            if isinstance(a, numbers.Real):
                res = getattr(math, self._name)(a)
            else:
                res = getattr(cmath, self._name)(a)
        except ValueError:
            warnings.warn(
                "Value error in evaluation of function %s with argument %s." % (self._name, a)
            )
            raise
        return res

    def __str__(self):
        """Format as a string."""
        return "%s(%s)" % (self._name, self.ufl_operands[0])


@ufl_type()
class Sqrt(MathFunction):
    """Square root."""

    __slots__ = ()

    def __new__(cls, argument):
        """Create a new Sqrt."""
        if isinstance(argument, (RealValue, Zero, numbers.Real)):
            if float(argument) < 0:
                return ComplexValue(cmath.sqrt(complex(argument)))
            else:
                return FloatValue(math.sqrt(float(argument)))
        if isinstance(argument, (ComplexValue, complex)):
            return ComplexValue(cmath.sqrt(complex(argument)))
        return MathFunction.__new__(cls)

    def __init__(self, argument):
        """Initialise."""
        MathFunction.__init__(self, "sqrt", argument)


@ufl_type()
class Exp(MathFunction):
    """Exponentiation.."""

    __slots__ = ()

    def __new__(cls, argument):
        """Create a new Exp."""
        if isinstance(argument, (RealValue, Zero)):
            return FloatValue(math.exp(float(argument)))
        if isinstance(argument, (ComplexValue)):
            return ComplexValue(cmath.exp(complex(argument)))
        return MathFunction.__new__(cls)

    def __init__(self, argument):
        """Initialise."""
        MathFunction.__init__(self, "exp", argument)


@ufl_type()
class Ln(MathFunction):
    """Natural logarithm."""

    __slots__ = ()

    def __new__(cls, argument):
        """Create a new Ln."""
        if isinstance(argument, (RealValue, Zero)):
            return FloatValue(math.log(float(argument)))
        if isinstance(argument, (ComplexValue)):
            return ComplexValue(cmath.log(complex(argument)))
        return MathFunction.__new__(cls)

    def __init__(self, argument):
        """Initialise."""
        MathFunction.__init__(self, "ln", argument)

    def evaluate(self, x, mapping, component, index_values):
        """Evaluate."""
        a = self.ufl_operands[0].evaluate(x, mapping, component, index_values)
        try:
            return math.log(a)
        except TypeError:
            return cmath.log(a)


@ufl_type()
class Cos(MathFunction):
    """Cosine."""

    __slots__ = ()

    def __new__(cls, argument):
        """Create a new Cos."""
        if isinstance(argument, (RealValue, Zero)):
            return FloatValue(math.cos(float(argument)))
        if isinstance(argument, (ComplexValue)):
            return ComplexValue(cmath.cos(complex(argument)))
        return MathFunction.__new__(cls)

    def __init__(self, argument):
        """Initialise."""
        MathFunction.__init__(self, "cos", argument)


@ufl_type()
class Sin(MathFunction):
    """Sine."""

    __slots__ = ()

    def __new__(cls, argument):
        """Create a new Sin."""
        if isinstance(argument, (RealValue, Zero)):
            return FloatValue(math.sin(float(argument)))
        if isinstance(argument, (ComplexValue)):
            return ComplexValue(cmath.sin(complex(argument)))
        return MathFunction.__new__(cls)

    def __init__(self, argument):
        """Initialise."""
        MathFunction.__init__(self, "sin", argument)


@ufl_type()
class Tan(MathFunction):
    """Tangent."""

    __slots__ = ()

    def __new__(cls, argument):
        """Create a new Tan."""
        if isinstance(argument, (RealValue, Zero)):
            return FloatValue(math.tan(float(argument)))
        if isinstance(argument, (ComplexValue)):
            return ComplexValue(cmath.tan(complex(argument)))
        return MathFunction.__new__(cls)

    def __init__(self, argument):
        """Initialise."""
        MathFunction.__init__(self, "tan", argument)


@ufl_type()
class Cosh(MathFunction):
    """Hyperbolic cosine."""

    __slots__ = ()

    def __new__(cls, argument):
        """Create a new Cosh."""
        if isinstance(argument, (RealValue, Zero)):
            return FloatValue(math.cosh(float(argument)))
        if isinstance(argument, (ComplexValue)):
            return ComplexValue(cmath.cosh(complex(argument)))
        return MathFunction.__new__(cls)

    def __init__(self, argument):
        """Initialise."""
        MathFunction.__init__(self, "cosh", argument)


@ufl_type()
class Sinh(MathFunction):
    """Hyperbolic sine."""

    __slots__ = ()

    def __new__(cls, argument):
        """Create a new Sinh."""
        if isinstance(argument, (RealValue, Zero)):
            return FloatValue(math.sinh(float(argument)))
        if isinstance(argument, (ComplexValue)):
            return ComplexValue(cmath.sinh(complex(argument)))
        return MathFunction.__new__(cls)

    def __init__(self, argument):
        """Initialise."""
        MathFunction.__init__(self, "sinh", argument)


@ufl_type()
class Tanh(MathFunction):
    """Hyperbolic tangent."""

    __slots__ = ()

    def __new__(cls, argument):
        """Create a new Tanh."""
        if isinstance(argument, (RealValue, Zero)):
            return FloatValue(math.tanh(float(argument)))
        if isinstance(argument, (ComplexValue)):
            return ComplexValue(cmath.tanh(complex(argument)))
        return MathFunction.__new__(cls)

    def __init__(self, argument):
        """Initialise."""
        MathFunction.__init__(self, "tanh", argument)


@ufl_type()
class Acos(MathFunction):
    """Inverse cosine."""

    __slots__ = ()

    def __new__(cls, argument):
        """Create a new Acos."""
        if isinstance(argument, (RealValue, Zero)):
            return FloatValue(math.acos(float(argument)))
        if isinstance(argument, (ComplexValue)):
            return ComplexValue(cmath.acos(complex(argument)))
        return MathFunction.__new__(cls)

    def __init__(self, argument):
        """Initialise."""
        MathFunction.__init__(self, "acos", argument)


@ufl_type()
class Asin(MathFunction):
    """Inverse sine."""

    __slots__ = ()

    def __new__(cls, argument):
        """Create a new Asin."""
        if isinstance(argument, (RealValue, Zero)):
            return FloatValue(math.asin(float(argument)))
        if isinstance(argument, (ComplexValue)):
            return ComplexValue(cmath.asin(complex(argument)))
        return MathFunction.__new__(cls)

    def __init__(self, argument):
        """Initialise."""
        MathFunction.__init__(self, "asin", argument)


@ufl_type()
class Atan(MathFunction):
    """Inverse tangent."""

    __slots__ = ()

    def __new__(cls, argument):
        """Create a new Atan."""
        if isinstance(argument, (RealValue, Zero)):
            return FloatValue(math.atan(float(argument)))
        if isinstance(argument, (ComplexValue)):
            return ComplexValue(cmath.atan(complex(argument)))
        return MathFunction.__new__(cls)

    def __init__(self, argument):
        """Initialise."""
        MathFunction.__init__(self, "atan", argument)


@ufl_type(is_scalar=True, num_ops=2)
class Atan2(Operator):
    """Inverse tangent with two inputs."""

    __slots__ = ()

    def __new__(cls, arg1, arg2):
        """Create a new Atan2."""
        if isinstance(arg1, (RealValue, Zero)) and isinstance(arg2, (RealValue, Zero)):
            return FloatValue(math.atan2(float(arg1), float(arg2)))
        if isinstance(arg1, (ComplexValue)) or isinstance(arg2, (ComplexValue)):
            raise TypeError("Atan2 does not support complex numbers.")
        return Operator.__new__(cls)

    def __init__(self, arg1, arg2):
        """Initialise."""
        Operator.__init__(self, (arg1, arg2))
        if isinstance(arg1, (ComplexValue, complex)) or isinstance(arg2, (ComplexValue, complex)):
            raise TypeError("Atan2 does not support complex numbers.")
        if not is_true_ufl_scalar(arg1):
            raise ValueError("Expecting scalar argument 1.")
        if not is_true_ufl_scalar(arg2):
            raise ValueError("Expecting scalar argument 2.")

    def evaluate(self, x, mapping, component, index_values):
        """Evaluate."""
        a = self.ufl_operands[0].evaluate(x, mapping, component, index_values)
        b = self.ufl_operands[1].evaluate(x, mapping, component, index_values)
        try:
            res = math.atan2(a, b)
        except TypeError:
            raise ValueError("Atan2 does not support complex numbers.")
        except ValueError:
            warnings.warn(
                "Value error in evaluation of function atan2 with arguments %s, %s." % (a, b)
            )
            raise
        return res

    def __str__(self):
        """Format as a string."""
        return "atan2(%s,%s)" % (self.ufl_operands[0], self.ufl_operands[1])


@ufl_type()
class Erf(MathFunction):
    """Erf function."""

    __slots__ = ()

    def __new__(cls, argument):
        """Create a new Erf."""
        if isinstance(argument, (RealValue, Zero)):
            return FloatValue(math.erf(float(argument)))
        if isinstance(argument, (ConstantValue)):
            return ComplexValue(math.erf(complex(argument)))
        return MathFunction.__new__(cls)

    def __init__(self, argument):
        """Initialise."""
        MathFunction.__init__(self, "erf", argument)

    def evaluate(self, x, mapping, component, index_values):
        """Evaluate."""
        a = self.ufl_operands[0].evaluate(x, mapping, component, index_values)
        return math.erf(a)


@ufl_type(is_abstract=True, is_scalar=True, num_ops=2)
class BesselFunction(Operator):
    """Base class for all bessel functions."""

    __slots__ = "_name"

    def __init__(self, name, nu, argument):
        """Initialise."""
        if not is_true_ufl_scalar(nu):
            raise ValueError("Expecting scalar nu.")
        if not is_true_ufl_scalar(argument):
            raise ValueError("Expecting scalar argument.")

        # Use integer representation if suitable
        fnu = float(nu)
        inu = int(nu)
        if fnu == inu:
            nu = as_ufl(inu)
        else:
            nu = as_ufl(fnu)

        Operator.__init__(self, (nu, argument))

        self._name = name

    def evaluate(self, x, mapping, component, index_values):
        """Evaluate."""
        a = self.ufl_operands[1].evaluate(x, mapping, component, index_values)
        try:
            import scipy.special
        except ImportError:
            raise ValueError(
                "You must have scipy installed to evaluate bessel functions in python."
            )
        name = self._name[-1]
        if isinstance(self.ufl_operands[0], IntValue):
            nu = int(self.ufl_operands[0])
            functype = "n" if name != "i" else "v"
        else:
            nu = self.ufl_operands[0].evaluate(x, mapping, component, index_values)
            functype = "v"
        func = getattr(scipy.special, name + functype)
        return func(nu, a)

    def __str__(self):
        """Format as a string."""
        return "%s(%s, %s)" % (self._name, self.ufl_operands[0], self.ufl_operands[1])


@ufl_type()
class BesselJ(BesselFunction):
    """Bessel J function."""

    __slots__ = ()

    def __init__(self, nu, argument):
        """Initialise."""
        BesselFunction.__init__(self, "cyl_bessel_j", nu, argument)


@ufl_type()
class BesselY(BesselFunction):
    """Bessel Y function."""

    __slots__ = ()

    def __init__(self, nu, argument):
        """Initialise."""
        BesselFunction.__init__(self, "cyl_bessel_y", nu, argument)


@ufl_type()
class BesselI(BesselFunction):
    """Bessel I function."""

    __slots__ = ()

    def __init__(self, nu, argument):
        """Initialise."""
        BesselFunction.__init__(self, "cyl_bessel_i", nu, argument)


@ufl_type()
class BesselK(BesselFunction):
    """Bessel K function."""

    __slots__ = ()

    def __init__(self, nu, argument):
        """Initialise."""
        BesselFunction.__init__(self, "cyl_bessel_k", nu, argument)



__authors__ = "Martin Sandve Alnæs"
__date__ = "2009-02-17 -- 2014-10-14"

import pytest

from ufl import (
    Coefficient,
    FunctionSpace,
    Mesh,
    SpatialCoordinate,
    as_vector,
    atan,
    cos,
    diff,
    exp,
    indices,
    ln,
    sin,
    tan,
    triangle,
    variable,
)
from ufl.algorithms import expand_derivatives
from ufl.constantvalue import as_ufl
from ufl.finiteelement import FiniteElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1


def get_variables():
    xv = None
    vv = 5.0
    return (xv, variable(vv))


@pytest.fixture
def v():
    xv, vv = get_variables()
    return vv


def _test(f, df):
    x, v = get_variables()

    dfv1 = diff(f(v), v)
    dfv2 = df(v)
    dfv1 = dfv1(x)
    dfv2 = dfv2(x)
    assert round(dfv1 - dfv2, 7) == 0

    dfv1 = diff(f(7 * v), v)
    dfv2 = 7 * df(7 * v)
    dfv1 = dfv1(x)
    dfv2 = dfv2(x)
    assert round(dfv1 - dfv2, 7) == 0


def testVariable(v):
    def f(v):
        return v

    def df(v):
        return as_ufl(1)

    _test(f, df)


def testSum(v):
    def f(v):
        return v + 1

    def df(v):
        return as_ufl(1)

    _test(f, df)


def testProduct(v):
    def f(v):
        return 3 * v

    def df(v):
        return as_ufl(3)

    _test(f, df)


def testPower(v):
    def f(v):
        return v**3

    def df(v):
        return 3 * v**2

    _test(f, df)


def testDivision(v):
    def f(v):
        return v / 3.0

    def df(v):
        return as_ufl(1.0 / 3.0)

    _test(f, df)


def testDivision2(v):
    def f(v):
        return 3.0 / v

    def df(v):
        return -3.0 / v**2

    _test(f, df)


def testExp(v):
    def f(v):
        return exp(v)

    def df(v):
        return exp(v)

    _test(f, df)


def testLn(v):
    def f(v):
        return ln(v)

    def df(v):
        return 1.0 / v

    _test(f, df)


def testSin(v):
    def f(v):
        return sin(v)

    def df(v):
        return cos(v)

    _test(f, df)


def testCos(v):
    def f(v):
        return cos(v)

    def df(v):
        return -sin(v)

    _test(f, df)


def testTan(v):
    def f(v):
        return tan(v)

    def df(v):
        return 2.0 / (cos(2.0 * v) + 1.0)

    _test(f, df)


# TODO: Check the following tests. They run into strange math domain errors.
# def testAsin(v):
#    def f(v):  return asin(v)
#    def df(v): return 1/sqrt(1.0 - v**2)
#    _test(f, df)

# def testAcos(v):
#    def f(v):  return acos(v)
#    def df(v): return -1/sqrt(1.0 - v**2)
#    _test(f, df)


def testAtan(v):
    def f(v):
        return atan(v)

    def df(v):
        return 1 / (1.0 + v**2)

    _test(f, df)


def testIndexSum(v):
    def f(v):
        # 3*v + 4*v**2 + 5*v**3
        a = as_vector((v, v**2, v**3))
        b = as_vector((3, 4, 5))
        (i,) = indices(1)
        return a[i] * b[i]

    def df(v):
        return 3 + 4 * 2 * v + 5 * 3 * v**2

    _test(f, df)


def testCoefficient():
    coord_elem = FiniteElement("Lagrange", triangle, 1, (3,), identity_pullback, H1)
    mesh = Mesh(coord_elem)
    V = FunctionSpace(mesh, FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1))
    v = Coefficient(V)
    assert round(expand_derivatives(diff(v, v)) - 1.0, 7) == 0


def testDiffX():
    cell = triangle
    domain = Mesh(FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1))
    x = SpatialCoordinate(domain)
    f = x[0] ** 2 * x[1] ** 2
    (i,) = indices(1)
    df1 = diff(f, x)
    df2 = as_vector(f.dx(i), i)

    xv = (2, 3)
    df10 = df1[0](xv)
    df11 = df1[1](xv)
    df20 = df2[0](xv)
    df21 = df2[1](xv)
    assert round(df10 - df20, 7) == 0
    assert round(df11 - df21, 7) == 0
    assert round(df10 - 2 * 2 * 9, 7) == 0
    assert round(df11 - 2 * 4 * 3, 7) == 0


# TODO: More tests involving wrapper types and indices



"""External operator.

This module defines the ``ExternalOperator`` class, which symbolically
represents operators that are not straightforwardly expressible in UFL.
Subclasses of ``ExternalOperator`` must define how this operator should
be evaluated as well as its derivatives from a given set of operands.
"""
# Copyright (C) 2019 Nacime Bouziani
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Nacime Bouziani, 2023

from ufl.core.base_form_operator import BaseFormOperator
from ufl.core.ufl_type import ufl_type


@ufl_type(num_ops="varying", is_differential=True)
class ExternalOperator(BaseFormOperator):
    """External operator."""

    # Slots are disabled here because they cause trouble in PyDOLFIN
    # multiple inheritance pattern:
    _ufl_noslots_ = True

    def __init__(self, *operands, function_space, derivatives=None, argument_slots=()):
        """Initialise.

        Args:
            operands: operands on which acts the ExternalOperator.
            function_space: the FunctionSpace, or MixedFunctionSpace on
                which to build this Function.
            derivatives: tuple specifying the derivative multiindex.
                argument_slots: tuple composed containing expressions with
                ufl.Argument or ufl.Coefficient objects.
            argument_slots: TODO
        """
        # -- Derivatives -- #
        if derivatives is not None:
            if not isinstance(derivatives, tuple):
                raise TypeError(f"Expecting a tuple for derivatives and not {derivatives}.")
            if not len(derivatives) == len(operands):
                raise ValueError(f"Expecting a size of {len(operands)} for {derivatives}.")
            if not all(isinstance(d, int) for d in derivatives) or any(d < 0 for d in derivatives):
                raise ValueError(
                    "Expecting a derivative multi-index with nonnegative indices, "
                    f"not {derivatives}."
                )
        else:
            derivatives = (0,) * len(operands)

        BaseFormOperator.__init__(
            self,
            *operands,
            function_space=function_space,
            derivatives=derivatives,
            argument_slots=argument_slots,
        )

    def ufl_element(self):
        """Shortcut to get the finite element of the function space of the external operator."""
        # Useful when applying split on an ExternalOperator
        return self.arguments()[0].ufl_element()

    def grad(self):
        """Returns the symbolic grad of the external operator."""
        # By default, differential rules produce `grad(assembled_o)`
        # `where assembled_o` is the `Coefficient` resulting from
        # assembling the external operator since the external operator
        # may not be smooth enough for chain rule to hold. Symbolic
        # gradient (`grad(ExternalOperator)`) depends on the operator
        # considered and its implementation may be needed in some cases
        # (e.g. convolution operator).
        raise NotImplementedError(
            "Symbolic gradient not defined for the external operator considered."
        )

    def assemble(self, *args, **kwargs):
        """Assemble the external operator."""
        raise NotImplementedError(
            f"Symbolic evaluation of {self._ufl_class_.__name__} not available."
        )

    def _ufl_expr_reconstruct_(
        self, *operands, function_space=None, derivatives=None, argument_slots=None, add_kwargs={}
    ):
        """Return a new object of the same type with new operands."""
        return type(self)(
            *operands,
            function_space=function_space or self.ufl_function_space(),
            derivatives=derivatives or self.derivatives,
            argument_slots=argument_slots or self.argument_slots(),
            **add_kwargs,
        )

    def __str__(self):
        """Default str string for ExternalOperator operators."""
        d = "\N{PARTIAL DIFFERENTIAL}"
        derivatives = self.derivatives
        d_ops = "".join(d + "o" + str(i + 1) for i, di in enumerate(derivatives) for j in range(di))
        e = "e("
        e += ", ".join(str(op) for op in self.ufl_operands)
        e += "; "
        e += ", ".join(str(arg) for arg in reversed(self.argument_slots()))
        e += ")"
        return d + e + "/" + d_ops if sum(derivatives) > 0 else e

    def __eq__(self, other):
        """Check for equality."""
        if self is other:
            return True
        return (
            type(self) is type(other)
            and all(a == b for a, b in zip(self.ufl_operands, other.ufl_operands))
            and all(a == b for a, b in zip(self._argument_slots, other._argument_slots))
            and self.derivatives == other.derivatives
            and self.ufl_function_space() == other.ufl_function_space()
        )



"""Mixin class for types with a global unique counter attached to each object."""

# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

import itertools


class Counted:
    """Mixin class for globally counted objects."""

    # Mixin classes do not work well with __slots__ so _count must be
    # added to the __slots__ of the inheriting class
    __slots__ = ()

    _counter = None

    def __init__(self, count=None, counted_class=None):
        """Initialize the Counted instance.

        Args:
            count: The object count, if ``None`` defaults to the next value
                according to the global counter (per type).
            counted_class: Class to attach the global counter too. If ``None``
                then ``type(self)`` will be used.

        """
        # create a new counter for each subclass
        counted_class = counted_class or type(self)
        if counted_class._counter is None:
            counted_class._counter = itertools.count()

        self._count = count if count is not None else next(counted_class._counter)
        self._counted_class = counted_class

    def count(self):
        """Get count."""
        return self._count
